(*****************************************************************************
 * HOL-TestGen --- theorem-prover based test case generation
 *                 http://www.brucker.ch/projects/hol-testgen/
 *                                                                            
 * hoare_test_theory.thy --- testing hoare-like specifications 
 * This file is part of HOL-TestGen.
 *
 * Copyright (c) 2005-2007, ETH Zurich, Switzerland
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)
(* $Id: hoare_test_theory.thy 7260 2008-02-10 22:27:50Z brucker $ *)

(* ML {* add_path "$ISABELLE_HOME/src/HOL/IMP"; *} *)
theory 
  hoare_test_theory 
imports
  VC
  Testing
begin

section{* Alternative Formulations of Hoare-Rules *}
lemma semi_rev: "\<lbrakk> |- {Q} d {R}; |- {P} c {Q} \<rbrakk> \<Longrightarrow> |- {P} c; d {R}"
  by(rule semi, assumption+)


lemma conseq_rev:
  "\<lbrakk>|- {P} c {Q}; \<forall>s. P' s \<longrightarrow> P s; \<forall>s. Q s \<longrightarrow> Q' s\<rbrakk> \<Longrightarrow> |- {P'} c {Q'}"
  by(rule conseq, auto)


lemma conseq_skip: 
      " (\<forall> s. P s \<longrightarrow> Q s) \<Longrightarrow>  |- {P} SKIP {Q}"
  by (rule conseq_rev, rule skip, auto)


lemma conseq_ass: 
      " (\<forall> s. P s \<longrightarrow> Q (s[x \<mapsto> E s])) \<Longrightarrow>  |- {P} x :== E {Q}"
  by (rule conseq_rev, rule ass, auto)


thm While

lemma conseq_while: 
      "\<lbrakk>  |- {\<lambda> s. I s \<and> cond s} s {I};
          \<forall>s. P s \<longrightarrow> I s;
          \<forall>s. I s \<and> \<not> cond s \<longrightarrow> Q s \<rbrakk>
       \<Longrightarrow>  |- {P} WHILE cond DO s {Q}"
  by (rule conseq_rev, rule While, assumption+)


section{* Hoare-Rules Using Semantic Equivalences *}
text{* Soundness and Completeness justify the following 
  equivalence, which allows to conclude the existence
  of Hoare-proofs from denotational-semantic equivalences. *} 

lemma hoare_deriv_eq_valid: "( |- {P} c {Q}) = ( |= {P} c {Q})"
  by(auto intro!: hoare_sound hoare_relative_complete)


lemma hoare_false: 
      " |- {\<lambda> x. False} c {Q}"
  by(simp add: hoare_deriv_eq_valid  hoare_valid_def)
  

lemma while_unfold: 
      "( |- {P} WHILE b DO c {Q}) = 
       ( |- {P}IF b THEN c; WHILE b DO c ELSE SKIP {Q})"
  by(simp only: hoare_deriv_eq_valid  hoare_valid_def 
                Denotation.C_While_If[symmetric])


lemma semi_skip1: 
      "( |- {P} SKIP ;c {Q}) = ( |- {P} c {Q})"
  by(simp only: hoare_deriv_eq_valid  hoare_valid_def 
                Denotation.C.simps Id_O_R R_O_Id)


lemma semi_skip2: 
      "( |- {P} c;SKIP {Q}) = ( |- {P} c {Q})"
  by(simp only: hoare_deriv_eq_valid hoare_valid_def 
                Denotation.C.simps Id_O_R R_O_Id)


lemma semi_assoc: 
      "( |- {P} (c;d) ;e {Q}) = ( |- {P} c; (d;e) {Q})"
  by(simp only: hoare_deriv_eq_valid hoare_valid_def 
                Denotation.C.simps Relation.O_assoc)


lemma semi_assoc2: 
      "( |- {P} (c;(d;e)) ; f {Q}) = ( |- {P} c;(d;(e;f)) {Q})"
  by(simp only: hoare_deriv_eq_valid hoare_valid_def 
                Denotation.C.simps Relation.O_assoc)


section{* Unwind and its Correctness *}
text{* The core of our white box testing function
   is the following ``unwind'' function, that
   ``unfolds'' while loops and normalizes the resulting
   program in order to expose it to the operational
   semantics (i.e. the ``natural semantics'' 
   @{text evalc} up to an unwind factor k. Evaluating 
   programs leads to accumulating  path-conditions: If a 
   remaining constraint (whose components essentially result 
   from applications of the @{text If_split} rule), is satisfiable that a 
   path through a program is traceable and results to a certain
   successor state.
   
   This can be used to test program specifications:
   Hoare-Triples were checked against for all paths up to 
   a certain depth.
*}


 
consts "@@" :: "[com,com] \<Rightarrow> com" (infixr 70)
primrec 
ca_skip : "SKIP @@ c = c"
ca_ass  : "(x:== E) @@ c = ((x:== E); c)"
ca_semi : "(c;d) @@ e = (c; d @@ e)"
ca_If   : "(IF b THEN c ELSE d) @@ e = 
                     (IF b THEN c @@ e ELSE d @@ e)"
ca_while: "(WHILE b DO c) @@ e = ((WHILE b DO c);e)"


lemma C_skip_cancel1[simp] : "C(SKIP;c) = C(c)"
  by (simp add: Denotation.C.simps Id_O_R R_O_Id)

lemma C_skip_cancel2[simp] : "C(c;SKIP) = C(c)"
  by (simp add: Denotation.C.simps Id_O_R R_O_Id)

lemma C_If_semi[simp] : 
      "C((IF x THEN c ELSE d);e) = C(IF x THEN (c;e) ELSE (d;e))"
  by auto

lemma comappend_correct [simp]: "C(c @@ d) = C(c;d)"
  apply(induct "c")
  apply(simp_all only: C_If_semi ca_If)
  apply(simp_all add: Relation.O_assoc)
  done  

consts unwind :: "nat \<times> com \<Rightarrow> com"
recdef unwind "less_than <*lex*> measure(\<lambda> s. size s)"
uw_skip : "unwind(n, SKIP)     = SKIP"
uw_ass  : "unwind(n, a :== E)  = (a :== E)"
uw_If   : "unwind(n, IF b THEN c ELSE d) = 
                 IF b THEN unwind(n, c) ELSE unwind(n, d)"
uw_while: "unwind(n, WHILE b DO c) = 
           (if 0 < n 
            then IF b THEN unwind(n,c)@@unwind(n- 1,WHILE b DO c) ELSE SKIP
            else WHILE b DO unwind(0, c))"
uw_semi1: "unwind(n, SKIP ; c) = unwind(n, c)"
uw_semi2: "unwind(n, c ; SKIP) = unwind(n, c)"
uw_semi3: "unwind(n, (IF b THEN c ELSE d) ; e) = 
                 (IF b THEN (unwind(n,c;e)) ELSE (unwind(n,d;e)))"
uw_semi4: "unwind(n, (c ; d); e) =  (unwind(n, c;d))@@(unwind(n,e))"
uw_semi5: "unwind(n, c ; d) =  (unwind(n, c))@@(unwind(n, d))"
(* Note that the Isabelle Recursion Package adopts a "First Fit"
   pattern matching strategy (similar SML). This means that in
   case of overlapping cases, the first is taken into account
   with higher priority --- this is reflected on the level of
   the rewrite rule set generated from this definition. 
   Thus, uw_semi4 is a catch-all rule for sequential composition. *)


lemma unwind_correct_aux1 :
  assumes H : "\<forall>x. C (unwind (x, c))   = C c"
  shows       "C(unwind(n,WHILE b DO c)) = C(WHILE b DO c)"
  apply(induct "n")
  apply(simp add: Denotation.C.simps H)
  apply(subst uw_while)
  apply(subst if_P, simp)
  apply(simp only: Denotation.C.simps  comappend_correct)
  apply(rule_tac s = "n" and t = "Suc n - 1" in subst)
  apply arith
  apply(simp only: Denotation.C.simps [symmetric] H)
  apply(simp only: Denotation.C_While_If)
  done


declare uw_while [simp del]

lemma unwind_correct_aux2 : 
  "C(unwind(n,c;d))= C(unwind(n,c) ; unwind(n, d))"
  apply (induct "c",   simp_all)
  apply (case_tac "d", simp_all)
  apply (case_tac "d", simp_all)
  apply (case_tac "d", simp_all)
  apply (simp_all only: C.simps[symmetric] C_If_semi)
  apply (case_tac "d", simp_all)
  done  


lemma unwind_correct : "C(unwind(n,c)) = C(c)"
  apply (rule_tac x = n in spec) (* generalization over n *)
  apply (induct "c")
  prefer 5
  apply (rule allI)
  apply (rule unwind_correct_aux1, simp, simp_all) 
  apply (case_tac "c1",case_tac "c2",
         simp_all add: unwind_correct_aux2)
  done




lemma unwind_hoare: 
      "( |- {P} unwind(n,c) {Q} ) = ( |- {P} c {Q})"
  by(simp only: unwind_correct hoare_deriv_eq_valid 
                hoare_valid_def)



lemma wp_unwind : "wp (c) (p) = wp(unwind(n,c)) (p) "
  by(simp add: wp_def unwind_correct)


lemma wp_test : "\<forall>\<sigma>. P \<sigma> \<longrightarrow> wp (unwind(k,c)) Q \<sigma> 
                 \<Longrightarrow> |- {P} c {Q}"
  apply  (rule Hoare.hoare_conseq1)
  prefer 2
  apply  (rule wp_is_pre)
  apply  (simp add: wp_unwind[symmetric])
  done


ML{*

fun thyp_ify n = EVERY[TestGen.mp_fy n,
                       TestGen.all_ify [] n,
                       rtac ((thm"THYP_def") RS (thm"HOL.meta_eq_to_obj_eq") RS (thm"iffD1")) n]
*}



section{* Relation to Operational Semantics *}

lemma If_split:
 "\<lbrakk> b s \<Longrightarrow> \<langle>c0,s\<rangle> \<longrightarrow>\<^sub>c s';
   \<not> b s \<Longrightarrow> \<langle>c1,s\<rangle> \<longrightarrow>\<^sub>c s' \<rbrakk>
  \<Longrightarrow> \<langle>\<IF> b \<THEN> c0 \<ELSE> c1,s\<rangle> \<longrightarrow>\<^sub>c s'"
  by (cases "b s", simp_all)

lemma If_splitE:
  "\<lbrakk>   \<langle>IF b THEN c ELSE d,s\<rangle> \<longrightarrow>\<^sub>c s';
       \<lbrakk> b s; \<langle>c,s\<rangle> \<longrightarrow>\<^sub>c s' \<rbrakk> \<Longrightarrow> P;
       \<lbrakk>  \<not> b s; \<langle>d,s\<rangle> \<longrightarrow>\<^sub>c s' \<rbrakk> \<Longrightarrow> P \<rbrakk> \<Longrightarrow> P"
  by(cases "b s", simp_all)
 
lemma setup_test : 
   "( |- {Pre} c {Post}) = 
      (\<forall>s t.  \<langle>unwind (n, c),s\<rangle> \<longrightarrow>\<^sub>c t \<longrightarrow> Pre s \<longrightarrow> Post t)"
  apply(subst unwind_hoare[symmetric])
  apply(subst hoare_deriv_eq_valid)
  apply(simp only: hoare_valid_def denotational_is_natural)
  done

consts STOP :: com
text{* uninterpreted constant stopping deduction 
       for cases reached for it \ldots *}

constdefs ASSERT :: "[bexp,com] \<Rightarrow> com" 
          "ASSERT b c \<equiv> IF b THEN c ELSE STOP"

constdefs AWHILE :: "[bexp,bexp,com] \<Rightarrow> com" 
          "AWHILE b invnt c \<equiv> ASSERT invnt (WHILE b DO (c; ASSERT invnt SKIP))"


end
 
