(*****************************************************************************
 * HOL-TestGen --- theorem-prover based test case generation
 *                 http://www.brucker.ch/projects/hol-testgen/
 *                                                                            
 * RBT_def.thy --- testing red-black trees
 * This file is part of HOL-TestGen.
 *
 * Copyright (c) 2005-2007, ETH Zurich, Switzerland
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)
(* $Id: RBT_def.thy 7263 2008-02-10 22:36:41Z brucker $ *)

 
header {* Defining Red-Black Trees *}


theory 
  RBT_def
imports
  Testing
begin




(******************************)

  (* mimicking ml's datatype order *)
text {*
 The implementation of Red-Black
trees is mainly based on the following datatype declaration:
*}
datatype ml_order = LESS | EQUAL | GREATER  

axclass ord_key < type      

consts
  compare :: "'a::ord_key \<Rightarrow> 'a \<Rightarrow> ml_order "  

axclass LINORDER < linorder, ord_key
  LINORDER_less    : "((compare x y) = LESS)    = (x < y)"
  LINORDER_equal   : "((compare x y) = EQUAL)   = (x = y)"
  LINORDER_greater : "((compare x y) = GREATER) = (y < x)"

(* defining red-black trees *)

types    'a item = "'a::ord_key" 

datatype  color = R | B

datatype 'a tree = E | T color "'a tree" "'a item" "'a tree"


  
(* invariants *)

text {*
In this example we have chosen not only to check if keys are stored or
deleted correctly in the trees but also to check if the trees fulfill
the balancing invariants. We formalize the red and black invariant by recursive
predicates:


*}

consts
  isin         :: "'a::LINORDER item \<Rightarrow> 'a tree \<Rightarrow> bool"
  isord        :: "('a::LINORDER item) tree \<Rightarrow> bool"
  redinv       :: "'a tree \<Rightarrow> bool"
  blackinv     :: "'a tree \<Rightarrow> bool"
  strong_redinv:: "'a tree \<Rightarrow> bool"
  max_B_height :: "'a tree \<Rightarrow> nat"

(* test for elements *)
primrec
  isin_empty : "isin x E = False"
  isin_branch: "isin x (T c a y b) = (((compare x y) = EQUAL) 
                                     | (isin x a) | (isin x b))"

(* test for order of elements *) 
primrec
  isord_empty : "isord E = True"
  isord_branch: "isord (T c a y b) 
                = (isord a \<and> isord b 
                \<and> (\<forall> x. isin x a \<longrightarrow> ((compare x y) = LESS))
                \<and> (\<forall> x. isin x b \<longrightarrow> ((compare x y) = GREATER)))"

(* weak red invariant: no red node has a red child *)  
recdef redinv "measure (%t. (size t))"
  redinv_1:   "redinv E = True"
  redinv_2:   "redinv (T B a y b) = (redinv a \<and> redinv b)"
  redinv_3:   "redinv (T R (T R a x b) y c) = False"
  redinv_4:   "redinv (T R a x (T R b y c)) = False"
  redinv_5:   "redinv (T R a x b) = (redinv a \<and> redinv b)"

(* strong red invariant: every red node has an immediate black ancestor, i.e. 
   the root is black and the weak red invariant holds *)
recdef strong_redinv "{}"
  Rinv_1:  "strong_redinv E = True"
  Rinv_2:  "strong_redinv (T R a y b) = False"
  Rinv_3:  "strong_redinv (T B a y b) = (redinv a \<and> redinv b)"

(* calculating maximal number of black nodes on any path from root to leaf *)
recdef max_B_height "measure (%t. (size t))"
  maxB_height_1:  "max_B_height E = 0"
  maxB_height_3:  "max_B_height (T B a y b) 
                   = Suc(max (max_B_height a) (max_B_height b))"
  maxB_height_2:  "max_B_height (T R a y b) 
                   = (max (max_B_height a) (max_B_height b))"

(* black invariant: number of black nodes equal on all paths from root to leaf *)  
recdef blackinv "measure (%t. (size t))"
  blackinv_1:  "blackinv E = True"
  blackinv_2:  "blackinv (T color a y b) 
               = ((blackinv a) \<and> (blackinv b) 
               \<and> ((max_B_height a) = (max_B_height b)))"
end
