(*****************************************************************************
 * HOL-TestGen --- theorem-prover based test case generation
 *                 http://www.brucker.ch/projects/hol-testgen/
 *                                                                            
 * Sequence_test.thy --- sequence testing 
 * This file is part of HOL-TestGen.
 *
 * Copyright (c) 2005-2007, ETH Zurich, Switzerland
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)
(* $Id: Sequence_test.thy 8103 2008-06-12 05:15:26Z brucker $ *)

header {* Sequence Testing *} 

theory Sequence_test 
imports
  List 
  Testing
begin

text{* 
  In this theory, we present present a simple reactive system and
  demonstrate and show how \testgen{} can be used for testing such
  systems.

  Our scenario is motivated by the following communcation scenario: A
  client sends a server a communication request and specifies a
  port-range $X$. The server non-deterministically chooses a port $Y$
  which is within the specified range. The client sends a sequence of
  data (abstracted away in our example to a constant @{term "Data"})
  on the port allocated by the server.  The communication is
  terminated by the client with a @{term "stop"} event. Using a
  \textsc{csp}-like notation, we can describe such a system 
  as follows:
  \begin{multline} 
    \mathit{req}?X \rightarrow \mathit{port}?Y[Y < X]\\ 
    \rightarrow
    (\mathit{rec}\;N \bullet \mathit{send}!D,Y \rightarrow
     \mathit{ack} \rightarrow N\; \Box\; \mathit{stop} \rightarrow
     \mathit{ack} \rightarrow \mathit{SKIP}) \nonumber 
  \end{multline}
  It is necessary for our approach that the protocol strictly
  alternates client-side and server-side events; thus, we will be able
  to construct in a test of the server a step-function ioprog (see
  below) that stimulates the server with an input and records its
  result. If a protocol does not have alternation in its events, it
  must be constructed by artificial acknowledge events; it is then a
  question of their generation in the test-harness, if they were sent
  anyway or if they correspond to something like ``server reacted
  within timebounds.''

  The \emph{stimulation sequence} of the system under test results
  just from the projection of this protocol to the input events:
  \begin{gather}
    \mathit{req}?X \rightarrow (\mathit{rec}\; N \bullet \mathit{send}!D,Y 
    \rightarrow N\;\Box\; \mathit{stop} \rightarrow \mathit{SKIP}) \nonumber
  \end{gather}
*}

section{* Basic Technique: Events with explicit variables *}

text{* 
  We define \emph{abstract traces} containing explicit variables $X$,
  $Y$, \ldots. The whole testcase-generation is done on the basis of
  the abstract traces. However, some small additional functions
  $substitute$ and $bind$ were used to replace them with concrete
  values during the run of the test-driver, as well as programs that
  check pre and post conditions on the concrete values occuring in the
  concrete run. 
*}

text{* 
  We specify explicit variables and a joined type containing abstract
  events (replacing values by explicit variables) as well as their
  concrete counterparts. 
*}

datatype vars = X | Y
datatype data = Data

datatype dvars = inp vars | outp vars

types    chan = int (* just to make it executable *)

datatype InEvent  = req chan       | reqA vars |
                    send data chan | sendA data vars |
                    stop

datatype OutEvent = port chan      | portA vars|
                    ack 

constdefs lookup :: "['a \<rightharpoonup> 'b, 'a] \<Rightarrow> 'b"
          "lookup env v \<equiv>  the(env v)"
          success:: "'\<alpha> option \<Rightarrow> bool"
          "success x \<equiv> case x of None => False | Some x => True"

types    event = "InEvent + OutEvent"

section{* The infra-structure of the observer: substitute and rebind *}

text{* 
  The predicate @{term "subsitute"} allows for mapping abstract events
  containing explicit variables to concrete events by substituting the
  variables by values communicated in the system run. It requires an
  environment (``substitution'') where the concrete values occuring in
  the system run were assigned to variables.  
*}
consts   substitute :: "[vars \<rightharpoonup> chan, InEvent] \<Rightarrow> InEvent"
primrec 
   "substitute env (req n)    = req n"   
   "substitute env (reqA v)   = req(lookup env v)"
   "substitute env (send d n) = send d n"
   "substitute env (sendA d v)= send d (lookup env v)"
   "substitute env stop       = stop"

text{* 
  The predicate @{term "rebind"} extracts from concrete output events
  the values and binds them to explicit variables in env. It should
  never be applied to abract values; therefore, we we can use an
  underspecified version (arbitrary). The predicate @{term "rebind"}
  only stores $?$-occurrences in the protocol into the environment;
  $!$-occurences were ignored.  Events, that are the same in the
  abstract as well as the concrete setting are treated as abstract
  event.
 
  In a way, @{term "rebind"} can be viewed as an abstraction of the
  concrete log produced at runtime.  
*}
consts   rebind :: "[vars \<rightharpoonup> chan, OutEvent] \<Rightarrow> vars \<rightharpoonup> chan"
primrec 
   "rebind env (port n)   = env(Y\<mapsto>n)"
   "rebind env ack        = env"


section{* Abstract Protocols and Abstract Stimulation Sequences *}

text{* 
  Now we encode the protocol automaton (abstract version) by a
  recursive acceptance predicate.  One could project the set of
  stimulation sequences just by filtering out the outEvents occuring
  in the traces.

  We will not pursue this approach since a more constructive
  presentation of the stimulation sequence set is advisable for
  testing.

  However, we show here how such concepts can be specified. 
*}

syntax  A :: "nat"
        B :: "nat"
        C :: "nat"
        D :: "nat"
        E :: "nat"

translations "A" == "0"
             "B" == "Suc A"
             "C" == "Suc B"
             "D" == "Suc C"
             "E" == "Suc D"

consts accept' :: "nat \<times> event list \<Rightarrow> bool"
recdef accept' "measure(\<lambda> (x,y). length y)"
   "accept'(A,(Inl(reqA X))#S)    = accept'(B,S)"
   "accept'(B,(Inr(portA Y))#S)   = accept'(C,S)"
   "accept'(C,(Inl(sendA d Y))#S) = accept'(D,S)"
   "accept'(D,(Inr(ack))#S)       = accept'(C,S)"
   "accept'(C,(Inl(stop))#S)      = accept'(E,S)"
   "accept'(E,[Inr(ack)])         = True"
   "accept'(x,y)                  = False"

constdefs accept :: "event list \<Rightarrow> bool"
   "accept s \<equiv> accept'(0,s)"

text{* 
  We proceed by modeling a subautomaton of the protocol automaton
  accept. 
*}

consts stim_trace' :: "nat \<times> InEvent list \<Rightarrow> bool"
recdef stim_trace' "measure(\<lambda> (x,y). length y)"
   "stim_trace'(A,(reqA X)#S)    = stim_trace'(C,S)"
   "stim_trace'(C,(sendA d Y)#S) = stim_trace'(C,S)"
   "stim_trace'(C,[stop])        = True"
   "stim_trace'(x,y)             = False"

constdefs stim_trace :: "InEvent list \<Rightarrow> bool"
   "stim_trace s \<equiv> stim_trace'(A,s)"


section{* The Post-Condition *}

consts
   postcond' :: "((vars \<rightharpoonup> int) \<times> unit \<times> InEvent \<times> OutEvent) \<Rightarrow> bool"

recdef postcond' "{}"
   "postcond' (env, x, req n, port m) = (m <= n)"
   "postcond' (env, x, send z n, ack) = (n = lookup env Y)"
   "postcond' (env, x, stop, ack)     = True"   
   "postcond' (env, x, y, z)          = False"

   
constdefs 
   postcond :: "(vars \<rightharpoonup> int) \<times> unit \<Rightarrow> InEvent \<Rightarrow> OutEvent \<Rightarrow> bool"
  "postcond  x y z \<equiv> postcond' (fst x, snd x, y, z)"

section{* Testing for successful system runs of the server under test *}

text{* 
  So far, we have not made any assumption on the state $\sigma'$ of
  the of our program under test ioprog. It could be a log of the
  actual system run. However, for simplicity, we give it inly a
  trivial state in this test specification.
*}

(*<*)
ML{* (* TestGen.trace_test:=true  *) *}
(*>*)

section{* Test-Generation: The Standard Approach *}

declare stim_trace_def[simp]
declare Mfold.simps[simp del]

test_spec "stim_trace trace \<longrightarrow> 
           success(Mfold trace (empty(X\<mapsto>init),()) 
                       (observer rebind substitute postcond ioprog))"
  apply(gen_test_cases 4 1 "ioprog")
  store_test_thm "reactive"

testgen_params [iterations=1000]
gen_test_data "reactive"

thm  reactive.test_data

testgen_params [
  gen_wrapper=false,
  setup_code="
    fun fst(x,y) = x
    fun snd(x,y) = y
  ",
  dataconv_code="
    fun sendToServer event Unity = let
      fun toServerData Data = server.Data
      fun toServerVars X    = server.X
        | toServerVars Y    = server.Y
      fun fromServerData server.Data = Data
      fun fromServerVars server.X    = X
        | fromServerVars server.Y    = Y
      fun convert_InEvent (req x)    = server.req x
        | convert_InEvent (reqA x)   = server.reqA (toServerVars x) 
        | convert_InEvent (send (x,y))  
                                = server.send  (toServerData x,y)
        | convert_InEvent (sendA (x,y)) 
                                = server.sendA (toServerData x, toServerVars y) 
        | convert_InEvent stop      = server.stop 
      fun convert_OutEvent (server.port x)  = Some(port x,Unity)
        | convert_OutEvent (server.portA x) = Some(portA (fromServerVars x),Unity)
        | convert_OutEvent server.ack       = Some(ack,Unity)
    in 
      convert_OutEvent (server.read (convert_InEvent event))
    end
   "]


gen_test_script "sequence_script.sml" reactive ioprog "sendToServer"

section{* Test-Generation: Refined Approach involving TP *}

text{* 
  An analysis of the previous approach shows, that random solving on
  trace patterns is obviously still quite ineffective. Although path
  coverage wrt. the input stimulation trace automaton can be achieved
  with a reasonable high probability, the depth remains limited.

  The idea is to produce a better test theorem by more specialized
  rules, that take the special form of the input stimulation protocol
  into account. 
*}

lemma start :
   "stim_trace'(A,x#S) = ((x = reqA X) \<and> stim_trace'(C,S))"
apply(cases "x", simp_all)
apply(rule_tac y="vars" in  vars.exhaust,simp_all)
done

lemma final[simp]:
"(stim_trace' (x, stop # S)) = ((x=C)\<and>(S=[]))"
apply(case_tac "x=Suc (Suc (A\<Colon>nat))", simp_all)
apply(cases "S",simp_all)
apply(case_tac "x=Suc (A\<Colon>nat)", simp_all)
apply(case_tac "x = (A\<Colon>nat)", simp_all)
apply(subgoal_tac "\<exists> xa. x = Suc(Suc(Suc xa))",erule exE,simp)
apply(arith)
done

lemma step1 :
   "stim_trace'(C,x#y#S) = ((x=sendA Data Y) \<and>stim_trace'(C,y#S))"
apply(cases "x", simp_all)
apply(rule_tac y="data" in  data.exhaust,simp_all)
apply(rule_tac y="vars" in  vars.exhaust,simp_all)
done

lemma step2: 
   "stim_trace'(C,[x]) = (x=stop)"
apply(cases "x", simp_all)
apply(rule_tac y="data" in  data.exhaust,simp_all)
apply(rule_tac y="vars" in  vars.exhaust,simp_all)
done

text{* 
  The three rules $start$, $step1$ and $step2$ give us a translation
  of a constraint of the form $\mathrm{stim_trace'}(x,[a,\ldots,b])$
  into a simple conjunction of equalities (in general: disjunction and
  existential quantifier will also occur). Since a formula of this
  form is an easy game for $fast_tac$ inside $gen\_test\_cases$, we
  will get dramatically better test theorems, where the constraints
  have been resolved already.  
*}

text{* We reconfigure the rewriter \ldots *}
declare start[simp] step1[simp] step2 [simp]


test_spec "stim_trace trace \<longrightarrow> 
           success(Mfold trace (empty(X\<mapsto>init),()) 
                    (observer rebind substitute postcond ioprog))"
  apply(gen_test_cases 40 1 "ioprog")
  store_test_thm "reactive2"

testgen_params [iterations=1]
gen_test_data "reactive2"

thm  reactive2.test_data

gen_test_script "sequence_script.sml" reactive2 ioprog "sendToServer"

text{* 
  Within the timeframe of 1 Minute, we get trace lengthes of about 40
  in the stimulation input protocol, which corresponds to traces of 80
  in the standard protocol. The examples shows, that it is not the
  length of traces that is a limiting factor of our approach. The main
  problem is the complexity in the stimulation automaton (size,
  brachning-factors, possible instantiations of parameter input).
*}
end
