theory PersonalFirewallDatatype
imports "../../src/PacketFilter/DatatypeTwoNets" "../../src/PacketFilter/PortCombinators" 
         "../../src/PacketFilter/PolicySimplification" "../../src/FWTesting"
begin


text{*

The most basic firewall scenario; there is a personal PC on one side and the Internet on the other. 
There are two policies: the first one allows all traffic from the PC to the Internet and denies
all coming into the PC. The second policy only allows specific ports from the PC. This scenario
comes in three variants: the first one specifies the allowed protocols directly, the second together 
with their respective port numbers, the third one only with the port numbers.

*}



text{*
Definitions of the subnets 
*}

constdefs
 
 PC :: "(DatatypeTwoNets subnet)"
"PC \<equiv> {{(a,b). a = one}}"

 Internet :: "DatatypeTwoNets subnet"
"Internet \<equiv> {{(a,b). a = two}}"


text{*
Definition of the testing constraint
*}

constdefs
 not_in_same_net :: "(DatatypeTwoNets,'b) packet \<Rightarrow> bool"
"not_in_same_net x \<equiv> (src x \<sqsubset> PC \<longrightarrow> dest x \<sqsubset> Internet) \<and> (src x \<sqsubset> Internet \<longrightarrow> dest x \<sqsubset> PC)" 

text {*
Definitions of the policies 

In fact, the short definitions wouldn't have to be written down - they are the automatically simplified versions of their big counterparts.

*}

constdefs
 strictPolicy :: "(DatatypeTwoNets,'b) Policy"
"strictPolicy \<equiv> deny_all ++ allow_all_from_to PC Internet"

 ProtocolPolicy :: "(DatatypeTwoNets,'b) Policy"
"ProtocolPolicy \<equiv> deny_all ++ allow_prots_from_to {http,ftp,imap} PC Internet"

 ProtocolPortPolicy :: "(DatatypeTwoNets,'b) Policy"
"ProtocolPortPolicy \<equiv> deny_all ++ allow_prots_from_port_to {(http,80::port),(ftp,24),(imap,21)} PC Internet"

 PortPolicy :: "(DatatypeTwoNets,'b) Policy"
"PortPolicy \<equiv> deny_all ++ allow_from_ports_to {80::port,24,21} PC Internet"

 ProtocolPolicyBig :: "(DatatypeTwoNets,'b) Policy"
"ProtocolPolicyBig \<equiv> deny_all ++ allow_prot_from_to http PC Internet ++ allow_prot_from_to ftp PC Internet++ allow_prot_from_to imap PC Internet"

 ProtocolPortPolicyBig :: "(DatatypeTwoNets,'b) Policy"
"ProtocolPortPolicyBig \<equiv> deny_all ++ (allow_prot_from_port_to http PC (80::port) Internet) ++ allow_prot_from_port_to ftp PC (24::port) Internet ++ allow_prot_from_port_to imap PC (21::port) Internet"

 PortPolicyBig :: "(DatatypeTwoNets,'b) Policy"
"PortPolicyBig \<equiv> deny_all ++ allow_from_port_to (80::port) PC Internet++ allow_from_port_to (24::port) PC Internet++ allow_from_port_to (21::port) PC Internet"

lemmas policyLemmas = strictPolicy_def ProtocolPolicy_def ProtocolPortPolicy_def PortPolicy_def PC_def Internet_def ProtocolPolicyBig_def ProtocolPortPolicyBig_def PortPolicyBig_def


text{*

Testing the three policies

*}


test_spec "not_in_same_net x \<longrightarrow> FUT x = strictPolicy x"
apply (prepareFWthm)
apply (simp add: policyLemmas stateless_rules DatatypeTwoNetsLemmas)
apply (gen_test_cases "FUT")
apply (simp_all add: not_in_same_net_def policyLemmas DatatypeTwoNetsLemmas)
store_test_thm "strict"

gen_test_data "strict"

test_spec "not_in_same_net x \<longrightarrow> FUT x = ProtocolPolicy x"
apply (prepareFWthm)
apply (simp add: policyLemmas stateless_rules DatatypeTwoNetsLemmas)
apply (gen_test_cases "FUT")
apply (simp_all add: not_in_same_net_def policyLemmas DatatypeTwoNetsLemmas)
store_test_thm "protocol"

gen_test_data "protocol"




(* or equivalently *)

test_spec "not_in_same_net x \<longrightarrow> FUT x = ProtocolPolicyBig x"
apply (simp add: ProtocolPolicyBig_def PSss)
apply (simplifyPolicy)
apply (tactic {*(((stac (thm "map_assoc")  1) THEN 
                 (asm_full_simp_tac (simpset () addsimps (thms "PolicySimplification")) 1)))*})


apply (tactic {*((stac (thm "map_assoc")  1) THEN 
                 (asm_full_simp_tac (simpset () Addsimps [thms "PolicySimplification"]) 1))*})

 
apply (subst map_assoc)
apply (simp add: PolicySimplification p2a1 p2a2)
apply (subst map_assoc)
apply (simp add: PolicySimplification p2a1 p2a2)
apply (subst map_assoc)

thm PolicySimplification
apply (subst PolicySimplification)

apply (prepareFWthm)
apply (simp add: policyLemmas stateless_rules DatatypeTwoNetsLemmas)
apply (gen_test_cases "FUT")
apply (simp_all add: not_in_same_net_def policyLemmas DatatypeTwoNetsLemmas)
store_test_thm "protocol"

gen_test_data "protocol"

*)

test_spec "not_in_same_net x \<longrightarrow> FUT x = ProtocolPortPolicy x"
apply (prepareFWthm)
apply (simp add: policyLemmas stateless_rules DatatypeTwoNetsLemmas PortCombinators)
apply (gen_test_cases "FUT")
apply (simp_all add: not_in_same_net_def policyLemmas DatatypeTwoNetsLemmas)
store_test_thm "protocolPort"


gen_test_data "protocolPort"

test_spec "not_in_same_net x \<longrightarrow> FUT x = PortPolicy x"
apply (prepareFWthm)
apply (simp add: policyLemmas stateless_rules DatatypeTwoNetsLemmas PortCombinators)
apply (gen_test_cases "FUT")
apply (simp_all add: not_in_same_net_def policyLemmas DatatypeTwoNetsLemmas)

store_test_thm "port"


gen_test_data "port"


(* BIG *)


test_spec "not_in_same_net x \<longrightarrow> FUT x = ProtocolPolicyBig x"
apply (prepareFWthm)
apply (simp add: policyLemmas stateless_rules DatatypeTwoNetsLemmas PortCombinators)
apply (gen_test_cases "FUT")
apply (simp_all add: not_in_same_net_def policyLemmas DatatypeTwoNetsLemmas)
store_test_thm "protocolBig"

gen_test_data "protocolBig"


test_spec "not_in_same_net x \<longrightarrow> FUT x = ProtocolPortPolicyBig x"
apply (prepareFWthm)
apply (simp add: policyLemmas stateless_rules DatatypeTwoNetsLemmas PortCombinators)
apply (gen_test_cases "FUT")
apply (simp_all add: not_in_same_net_def policyLemmas DatatypeTwoNetsLemmas)
store_test_thm "protocolPortBig"


gen_test_data "protocolPortBig"


test_spec "not_in_same_net x \<longrightarrow> FUT x = PortPolicyBig x"
apply (prepareFWthm)
apply (simp add: policyLemmas stateless_rules DatatypeTwoNetsLemmas PortCombinators allow_from_port_to_def)
apply (gen_test_cases "FUT")
apply (simp_all add: not_in_same_net_def policyLemmas DatatypeTwoNetsLemmas)

store_test_thm "portBig"


gen_test_data "portBig"





test_spec "not_in_same_net x \<longrightarrow> FUT x = strictPolicy x"
apply (prepareFWthm)
apply (simp add: policyLemmas stateless_rules DatatypeTwoNetsLemmas)
apply (gen_test_cases "FUT")
apply (simp_all add: not_in_same_net_def policyLemmas DatatypeTwoNetsLemmas)
store_test_thm "strict"

gen_test_data "strict"


end