(*****************************************************************************
 * HOL-TestGen --- theorem-prover based test case generation
 *                 http://www.brucker.ch/projects/hol-testgen/
 *                                                                            
 * --- 
 * This file is part of HOL-TestGen.
 *
 * Copyright (c) 2005-2008 ETH Zurich, Switzerland
 *               2008 Achim D. Brucker, Germany 
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)
(* $Id: DMZProtocolPort.thy 7470 2008-03-22 00:39:03Z brucker $ *)

theory 
  DMZProtocolPort
imports 
  IntegerPairNetwork 
  PortCombinators 
  FWTesting
begin

text{* This is the fourth scenario, slightly more complicated than the previous one, as we now also model specific servers within one network. Therefore, we cannot use anymore the modelling using datatypes, but only use the one where an address is modelled as an integer (with ports).

The scenario is the following:

\begin{labeling}{Networks:}
\item[Networks:]
  \begin{itemize}
  \item Intranet (Company intern network)
  \item DMZ (demilitarised zone, servers, etc), containing
     at least two distinct servers ``mail'' and ``web''
  \item Internet (``all others'') 
  \end{itemize}
\item[Policy:]
  \begin{itemize}
  \item allow http(s) from Intranet to Internet
  \item deny all trafic from Internet to Intranet
  \item allo imaps and smtp from intranet to mailserver
  \item allow smtp from Internet to mailserver
  \item allow http(s) from Internet to webserver 
  \item deny everything else
  \end{itemize}
  
\end{labeling}
*}


constdefs
   intranet::"IntegerPair subnet"
   "intranet \<equiv> {{(a,b) . (a > 1 \<and> a < 4) }}"

  dmz :: "IntegerPair subnet"
  "dmz \<equiv> {{(a,b). (a > 6) \<and> (a < 11)}}"
  mail :: "IntegerPair subnet"
  "mail \<equiv> {{(a,b). a = 7}}"
  web :: "IntegerPair subnet"
  "web \<equiv> {{(a,b). a = 8 }}"

  internet :: "IntegerPair subnet"
  "internet \<equiv> {{(a,b).  \<not> ( (a > 1 \<and> a < 4) \<or> (a > 6) \<and> (a < 11))   }}"

constdefs 
  Intranet_mail :: "(IntegerPair,'b) Rule"
  "Intranet_mail \<equiv> (allow_prots_from_to {smtp,imap} intranet mail)" 

  Intranet_Internet :: "(IntegerPair,'b) Rule"
  "Intranet_Internet \<equiv> allow_prots_from_to {http,https} intranet internet"
     
  Internet_web :: "(IntegerPair,'b) Rule"
  "Internet_web \<equiv> (allow_prots_from_to {http,https} internet web)"
          
  Internet_mail :: "(IntegerPair,'b) Rule"
  "Internet_mail \<equiv>  (allow_prot_from_to smtp internet dmz)"

constdefs 
  Intranet_mail_pp :: "(IntegerPair,'b) Rule"
  "Intranet_mail_pp \<equiv> (allow_prots_from_port_to {(smtp,21::port),(imap,24)} intranet mail)" 

  Intranet_Internet_pp :: "(IntegerPair,'b) Rule"
  "Intranet_Internet_pp \<equiv> allow_prots_from_port_to {(http,80::port),(https,90)} intranet internet"
     
  Internet_web_pp :: "(IntegerPair,'b) Rule"
  "Internet_web_pp \<equiv> (allow_prots_from_port_to {(http,80::port),(https,90)} internet web)"
          
  Internet_mail_pp :: "(IntegerPair,'b) Rule"
  "Internet_mail_pp \<equiv>  (allow_prot_from_port_to smtp internet (21::port) dmz)"

constdefs 
  Intranet_mail_port :: "(IntegerPair,'b) Rule"
  "Intranet_mail_port \<equiv> (allow_from_ports_to {21::port,14} intranet mail)" 

  Intranet_Internet_port :: "(IntegerPair,'b) Rule"
  "Intranet_Internet_port \<equiv> allow_from_ports_to {80::port,90} intranet internet"
     
  Internet_web_port :: "(IntegerPair,'b) Rule"
  "Internet_web_port \<equiv> (allow_from_ports_to {80::port,90} internet web)"
          
  Internet_mail_port :: "(IntegerPair,'b) Rule"
  "Internet_mail_port \<equiv>  (allow_all_from_port_to internet (21::port) dmz)"



constdefs 
  policyProtocol :: "(IntegerPair, DummyContent) Policy"
 "policyProtocol \<equiv> deny_all ++ 
                  Intranet_Internet ++ 
                  Intranet_mail ++  
                  Internet_mail ++ 
                  Internet_web"

constdefs 
  policyProtocolPort :: "(IntegerPair, DummyContent) Policy"
 "policyProtocolPort \<equiv> deny_all ++ 
                  Intranet_Internet_pp ++ 
                  Intranet_mail_pp ++  
                  Internet_mail_pp ++ 
                  Internet_web_pp"

constdefs 
  policyPort :: "(IntegerPair, DummyContent) Policy"
 "policyPort \<equiv> deny_all ++ 
                  Intranet_Internet_port ++ 
                  Intranet_mail_port ++  
                  Internet_mail_port ++ 
                  Internet_web_port"


text {* We only want to create test cases which are sent between the three main networks --- e.g. not between the mailserver and the dmz. Therefore, the constraint looks as follows. *}

constdefs
 not_in_same_net :: "(IntegerPair,DummyContent) packet \<Rightarrow> bool"
 "not_in_same_net x  \<equiv>  (src x \<sqsubset> internet \<longrightarrow> \<not> dest x \<sqsubset> internet) \<and> 
                        (src x \<sqsubset> intranet \<longrightarrow> \<not> dest x \<sqsubset> intranet) \<and> 
                        (src x \<sqsubset> dmz \<longrightarrow> \<not> dest x \<sqsubset> dmz)"

lemmas PolicyLemmas =  policyProtocol_def policyPort_def policyProtocolPort_def dmz_def internet_def intranet_def mail_def web_def 
                       Intranet_Internet_def Intranet_mail_def 
                       Internet_web_def Internet_mail_def
   Intranet_Internet_pp_def Intranet_mail_pp_def 
                       Internet_web_pp_def Internet_mail_pp_def
   Intranet_Internet_port_def Intranet_mail_port_def 
                       Internet_web_port_def Internet_mail_port_def




test_spec "not_in_same_net x \<longrightarrow> FUT (x:: (IntegerPair,DummyContent) packet) = policyProtocolPort x"
apply (prepareFWthm)
apply (simp add: PolicyLemmas IntegerNetworkLemmas PortCombinators stateless_rules )
apply (gen_test_cases "FUT")

apply (simp_all add: not_in_same_net_def src_def dest_def in_subnet_def PolicyLemmas)
store_test_thm "test"

end
