(*****************************************************************************
 * HOL-TestGen --- theorem-prover based test case generation
 *                 http://www.brucker.ch/projects/hol-testgen/
 *                                                                            
 * --- 
 * This file is part of HOL-TestGen.
 *
 * Copyright (c) 2005-2008 ETH Zurich, Switzerland
 *               2008 Achim D. Brucker, Germany 
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)
(* $Id: PersonalFirewallIpv4.thy 7470 2008-03-22 00:39:03Z brucker $ *)

theory 
  PersonalFirewallIpv4
imports 
  IPv4
  PortCombinators 
  FWTesting
begin


text{*

The most basic firewall scenario; there is a personal PC on one side and the Internet on the other. 
There are two policies: the first one allows all traffic from the PC to the Internet and denies
all coming into the PC. The second policy only allows specific ports from the PC. This scenario
comes in three variants: the first one specifies the allowed protocols directly, the second together 
with their respective port numbers, the third one only with the port numbers.

*}



text{*
Definitions of the subnets 
*}

constdefs
 
 PC :: "(ipv4 subnet)"
"PC \<equiv> {{((a,b,c,d),e). a = 1 \<and> b = 3 \<and> c = 5 \<and> d = 2}}"

 Internet :: "ipv4 subnet"
"Internet \<equiv> {{((a,b,c,d),e). \<not> (a = 1 \<and> b = 3 \<and> c = 5 \<and> d = 2)}}"


text{*
Definition of the testing constraint
*}

constdefs
 not_in_same_net :: "(ipv4,'b) packet \<Rightarrow> bool"
"not_in_same_net x \<equiv> (src x \<sqsubset> PC \<longrightarrow> dest x \<sqsubset> Internet) \<and> (src x \<sqsubset> Internet \<longrightarrow> dest x \<sqsubset> PC)" 

text {*
Definitions of the policies 
*}

constdefs
 strictPolicy :: "(ipv4,'b) Policy"
"strictPolicy \<equiv> deny_all ++ allow_all_from_to PC Internet"

 ProtocolPolicy :: "(ipv4,'b) Policy"
"ProtocolPolicy \<equiv> deny_all ++ allow_prots_from_to {http,ftp,imap} PC Internet"

 ProtocolPortPolicy :: "(ipv4,'b) Policy"
"ProtocolPortPolicy \<equiv> deny_all ++ allow_prots_from_port_to {(http,80::port),(ftp,24),(imap,21)} PC Internet"

 PortPolicy :: "(ipv4,'b) Policy"
"PortPolicy \<equiv> deny_all ++ allow_from_ports_to {80::port,24,21} PC Internet"


 ProtocolPolicyBig :: "(ipv4,'b) Policy"
"ProtocolPolicyBig \<equiv> deny_all ++ allow_prot_from_to http PC Internet ++ allow_prot_from_to ftp PC Internet++ allow_prot_from_to imap PC Internet"

 ProtocolPortPolicyBig :: "(ipv4,'b) Policy"
"ProtocolPortPolicyBig \<equiv> deny_all ++ (allow_prot_from_port_to http PC (80::port) Internet) ++ allow_prot_from_port_to ftp PC (24::port) Internet ++ allow_prot_from_port_to imap PC (21::port) Internet"

 PortPolicyBig :: "(ipv4,'b) Policy"
"PortPolicyBig \<equiv> deny_all ++ allow_from_port_to (80::port) PC Internet++ allow_from_port_to (24::port) PC Internet++ allow_from_port_to (21::port) PC Internet"


lemmas policyLemmas = strictPolicy_def ProtocolPolicy_def ProtocolPortPolicy_def PortPolicy_def PC_def Internet_def
ProtocolPolicyBig_def ProtocolPortPolicyBig_def PortPolicyBig_def

text{*

Testing the three policies

*}


test_spec "not_in_same_net x \<longrightarrow> FUT x = strictPolicy x"
apply (prepareFWthm)
apply (simp add: policyLemmas stateless_rules IPv4Lemmas)
apply (gen_test_cases "FUT")
apply (simp_all add: not_in_same_net_def policyLemmas IPv4Lemmas)

store_test_thm "strict"

test_spec "not_in_same_net x \<longrightarrow> FUT x = ProtocolPolicy x"
apply (prepareFWthm)
apply (simp add: policyLemmas stateless_rules IPv4Lemmas)
apply (gen_test_cases "FUT")
apply (simp_all add: not_in_same_net_def policyLemmas IPv4Lemmas)
store_test_thm "protocol"

test_spec "not_in_same_net x \<longrightarrow> FUT x = ProtocolPortPolicy x"
apply (prepareFWthm)
apply (simp add: policyLemmas stateless_rules IPv4Lemmas PortCombinators)
apply (gen_test_cases "FUT")
apply (simp_all add: not_in_same_net_def policyLemmas IPv4Lemmas)
store_test_thm "protocolPort"


test_spec "not_in_same_net x \<longrightarrow> FUT x = PortPolicy x"
apply (prepareFWthm)
apply (simp add: policyLemmas stateless_rules IPv4Lemmas PortCombinators)
apply (gen_test_cases "FUT")
apply (simp_all add: not_in_same_net_def policyLemmas IPv4Lemmas)
store_test_thm "port"



(* BIG *)


test_spec "not_in_same_net x \<longrightarrow> FUT x = ProtocolPolicyBig x"
apply (prepareFWthm)
apply (simp add: policyLemmas stateless_rules IPv4Lemmas PortCombinators)
apply (gen_test_cases "FUT")
apply (simp_all add: not_in_same_net_def policyLemmas IPv4Lemmas)
store_test_thm "protocolBig"

test_spec "not_in_same_net x \<longrightarrow> FUT x = ProtocolPortPolicyBig x"
apply (prepareFWthm)
apply (simp add: policyLemmas stateless_rules IPv4Lemmas PortCombinators)
apply (gen_test_cases "FUT")
apply (simp_all add: not_in_same_net_def policyLemmas IPv4Lemmas)
store_test_thm "protocolPortBig"

test_spec "not_in_same_net x \<longrightarrow> FUT x = PortPolicyBig x"
apply (prepareFWthm)
apply (simp add: policyLemmas stateless_rules IPv4Lemmas PortCombinators allow_from_port_to_def)
apply (gen_test_cases "FUT")
apply (simp_all add: not_in_same_net_def policyLemmas IPv4Lemmas)

store_test_thm "portBig"


end
