(*****************************************************************************
 * HOL-TestGen --- theorem-prover based test case generation
 *                 http://www.brucker.ch/projects/hol-testgen/
 *                                                                            
 * --- 
 * This file is part of HOL-TestGen.
 *
 * Copyright (c) 2005-2008 ETH Zurich, Switzerland
 *               2008 Achim D. Brucker, Germany 
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)
(* $Id: SimpleDMZProtocolPort.thy 7470 2008-03-22 00:39:03Z brucker $ *)

theory 
  SimpleDMZProtocolPort
imports 
  IntegerPairNetwork
  PortCombinators
  FWTesting
begin


text{*

This is the Simple DMZ example. There are three subnetworks, with either none or some protocols allowed between then. 

The policy is always in its simplest form, but specified in three variants: 
  - The allowed protocols named
  - The allowed protocols named including their port number
  - The allowed protocols only specified through their port number


The used network model is the Integer model

*}



constdefs
  intranet::"IntegerPair subnet"
  "intranet \<equiv> {{(a,b) . (a > 2) \<and> (a < 5) }}"

  dmz :: "IntegerPair subnet"
  "dmz \<equiv> {{(a,b). (a > 6) \<and> (a < 9)}}"

  internet :: "IntegerPair subnet"
  "internet \<equiv> {{(a,b).  \<not> ( ((a > 2) \<and> (a < 5) ) \<or> ((a > 6) \<and> (a < 9)))   }}"


constdefs 
  Intranet_DMZ :: "(IntegerPair,'b) Rule"
  "Intranet_DMZ \<equiv> (allow_prots_from_to {smtp,imap} intranet dmz)"

  Intranet_Internet :: "(IntegerPair,'b) Rule"
  "Intranet_Internet \<equiv> allow_prot_from_to http intranet internet"
           
  Internet_DMZ :: "(IntegerPair,'b) Rule"
  "Internet_DMZ \<equiv> (allow_prots_from_to {http,smtp} internet dmz)"

constdefs 
  Intranet_DMZ_ProtocolPort :: "(IntegerPair,'b) Rule"
  "Intranet_DMZ_ProtocolPort \<equiv> (allow_prots_from_port_to {(smtp,21::port),(imap,24)} intranet dmz)"

  Intranet_Internet_ProtocolPort :: "(IntegerPair,'b) Rule"
  "Intranet_Internet_ProtocolPort \<equiv> allow_prot_from_port_to http intranet (80::port) internet"
           
  Internet_DMZ_ProtocolPort :: "(IntegerPair,'b) Rule"
  "Internet_DMZ_ProtocolPort \<equiv> (allow_prots_from_port_to {(http,80::port),(smtp,21)} internet dmz)"


constdefs 
  Intranet_DMZ_Port :: "(IntegerPair,'b) Rule"
  "Intranet_DMZ_Port \<equiv> (allow_from_ports_to {21::port,24} intranet dmz)"

  Intranet_Internet_Port :: "(IntegerPair,'b) Rule"
  "Intranet_Internet_Port \<equiv> allow_all_from_port_to intranet (80::port) internet"
           
  Internet_DMZ_Port :: "(IntegerPair,'b) Rule"
  "Internet_DMZ_Port \<equiv> (allow_from_ports_to {80::port,21} internet dmz)"

  
constdefs 
  policyProtocol :: "(IntegerPair, DummyContent) Policy"
 "policyProtocol \<equiv> deny_all ++ 
                  Intranet_Internet ++
                  Intranet_DMZ ++  
                  Internet_DMZ"

constdefs 
  policyProtocolPort :: "(IntegerPair, DummyContent) Policy"
 "policyProtocolPort \<equiv> deny_all ++ 
                  Intranet_Internet_ProtocolPort  ++
                  Intranet_DMZ_ProtocolPort ++  
                  Internet_DMZ_ProtocolPort"

constdefs 
  policyPort :: "(IntegerPair, DummyContent) Policy"
 "policyPort \<equiv> deny_all ++ 
                  Intranet_Internet_Port ++
                  Intranet_DMZ_Port ++  
                  Internet_DMZ_Port"

lemmas PolicyLemmas =  policyProtocol_def dmz_def internet_def intranet_def
                       Intranet_Internet_def Intranet_DMZ_def 
                       Internet_DMZ_def dmz_def internet_def intranet_def
                       Intranet_Internet_ProtocolPort_def Intranet_DMZ_ProtocolPort_def 
                       Internet_DMZ_ProtocolPort_def policyProtocolPort_def
                       Intranet_Internet_Port_def Intranet_DMZ_Port_def 
                       Internet_DMZ_Port_def policyPort_def

constdefs
 not_in_same_net :: "(IntegerPair,'b) packet \<Rightarrow> bool"
 "not_in_same_net x  \<equiv>  (src x \<sqsubset> internet \<longrightarrow> \<not> dest x \<sqsubset> internet) \<and> 
                     (src x \<sqsubset> intranet \<longrightarrow> \<not> dest x \<sqsubset> intranet) \<and> 
                     (src x \<sqsubset> dmz \<longrightarrow> \<not> dest x \<sqsubset> dmz)"


test_spec "not_in_same_net x \<longrightarrow> FUT x = policyProtocolPort x"
apply (prepareFWthm)
apply (simp add: PolicyLemmas stateless_rules PortCombinators IntegerNetworkLemmas)
apply (gen_test_cases "FUT")
apply (simp_all add: not_in_same_net_def IntegerNetworkLemmas PolicyLemmas)  

store_test_thm "ProtocolPortPolicy"


end
