(*****************************************************************************
 * HOL-TestGen --- theorem-prover based test case generation
 *                 http://www.brucker.ch/projects/hol-testgen/
 *                                                                            
 * IPv4.thy --- modelling ipv4 addresses
 * This file is part of HOL-TestGen.
 *
 * Copyright (c) 2005-2007, ETH Zurich, Switzerland
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)
(* $Id: IPv4.thy 7266 2008-02-11 20:57:10Z brucker $ *)

header{* IPv4 *}
theory IPv4
imports NetworkCore
begin

section{* IPv4-Addresses *}
text{* 
  We defined an address as an abstract type to hold the whole model as
  generic as possible. But of course, when we want to have concrete
  test data, an address has to be concretized. In this theory we
  define IPv4 as one particular example of a representation of an
  address.
*}


text{*
  @{text "ipv4_ip"} is a common IP address. It is defined as a
  four-tupel of integers, to make them look almost like the common
  writing of these addresses (\texttt{217.82.1.1}
  vs. \texttt{(217,82,1,1)}). Technically it would be easier to define
  an address as only one integer, but for the sake of readability we
  chose this representation. For the same reason we didn't took
  integers instead of natural numbers, even though the numbers should
  of course never be negative.

  It should be modelled somehow that these numbers have to be smaller
  than 256. This has not been done yet but actually this doesn't make
  any problems so far as \testgen always picks low numbers in the
  random generation phase.

  A port is in our view also a part of an address. This gives us the
  possibility to use the model for a scenario where we either do not
  have or do not care about the port numbers. They are also defined as
  integers.  The full IPv4-address is then a tuple of the @{text
  "ipv4_ip"} and the @{text "port"}.  
*}







types 
  ipv4_ip = "(int \<times> int \<times> int \<times> int)" 
  port    =  "int"
  ipv4    =  "(ipv4_ip \<times> port)"

(*
instance port < port ..
*)

defs 
src_port_def: "src_port (x::(ipv4,'b) packet)  \<equiv> (snd o fst o snd o snd) x"
dest_port_def:"dest_port (x::(ipv4,'b) packet)  \<equiv> (snd o fst o snd o snd o snd) x"



text{* 
  As a help for later we also define constructors which take an
  address or an IP and return a subnet with all addresses having the
  same ipv4\_ip. 
*}

constdefs 
  subnet_of :: "ipv4  \<Rightarrow> ipv4 subnet"
  "subnet_of adr \<equiv> {{(a,b). (a = fst adr)}}"
 
  subnet_of_ip :: "ipv4_ip \<Rightarrow> ipv4 subnet"
  "subnet_of_ip ip \<equiv> {{(a,b). (a = ip)}}"

lemma in_subnet: 
  "((((a,b,c,d), e)) \<sqsubset> {{((x1,x2,x3,x4),y). P x1 x2 x3 x4 y}}) = (P a b c d e)"
  by (simp add: in_subnet_def)

lemma src_in_subnet: 
  "((src(q,w,((a,b,c,d),e),r,t)) \<sqsubset> {{((x1,x2,x3,x4),y). P x1 x2 x3 x4 y}}) = (P a b c d e)"
  by (simp add: in_subnet src_def)

lemma dest_in_subnet: 
  "((dest (q,w,r,((a,b,c,d),e),t)) \<sqsubset> {{((x1,x2,x3,x4),y). P x1 x2 x3 x4 y}}) = (P a b c d e)"
  by (simp add: in_subnet dest_def)

lemma protocol_is: "protocol (a,p,d,b,e) = p"
    by (simp add: protocol_def in_subnet)

lemma src_port_is: "src_port (a,c,(x::ipv4 adr),d,e) = snd x"
  by (simp add: src_port_def in_subnet)

lemma dest_port_is: "dest_port (a,c,d,(x::ipv4 adr),e) = snd x"
  by (simp add: dest_port_def in_subnet)



lemmas IPv4Lemmas = in_subnet src_in_subnet dest_in_subnet protocol_is src_port_is dest_port_is


end

