(*****************************************************************************
 * HOL-TestGen --- theorem-prover based test case generation
 *                 http://www.brucker.ch/projects/hol-testgen/
 *                                                                            
 * FWBasics.thy --- modelling firewalls 
 * This file is part of HOL-TestGen.
 *
 * Copyright (c) 2005-2007, ETH Zurich, Switzerland
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)
(* $Id: NetworkCore.thy 7299 2008-03-05 19:53:41Z lukasbru $ *)

header {* Firewall Basics *}
theory NetworkCore
imports Main
begin

text{* 
  A firewall operates on a network and observes the packets sent over
  it. We hence start with modelling these two things in this theory,
  trying to stay as generic as possible to make the model easy
  extendible.
*}

section{*  Packets *}

text{*
  Of all the fields of a real TCP/IP packet, we only want to model
  those which are necessary for the firewall to take its
  decision. Recalling the ruleset of Figure \ref{policy}, we see that
  we need the source and destination address and the application layer
  protocol. Ports, if needed, are modelled as part of the
  address. Additionally we include an ID and a content for later
  extensions.  
*}

text{*  The ID is an integer. *}
types id = int

text{* 
  The protocol is a datatype. Here we included the most common ones
  which were used for our case studies. If necessary, more application
  layer protocols could easily be added here.
*}

datatype protocol = ftp 
                  | http 
                  | https
                  | imap 
                  | imaps
                  | smtp 
                  | udp
                  | IP
                  | voip
                  | icmp
                  | unknown 
                  | tcp
                  | bgp
                  | pim
                  | igmp
                  | ntp
              
(*

datatype protocol = 
                   http | other
 
   *)            
text{* 
  We do not want to give a specific representation of an address yet
  in order to make it possible to have different representations
  (e.g. IPv4 and IPv6, with or without ports). To achieve this, we use
  type variables known from functional programming languages like
  Haskell\footnote{Type variables are usually denoted with greek
  letters, whereas in Isabelle they are written as @{text "'a"}.}. The
  source and the destination are both of the same type 
  @{text "'a adr"}.
*}
types 'a adr  = "'a"
         'a src  = "'a adr"
         'a dest = "'a adr"

text{* The content is also specified with a type variable. *}
types 'b content = "'b"

text {* 
  For applications were the concrete representation of the content field does 
  not matter, we provide a simple default type which can be used in that cases:
*}
 
datatype DummyContent = content


text{* 
  A packet then is a five-tuple consisting of the above declared
  types.
*}
types ('a,'b) packet = "id \<times> protocol \<times> 'a src \<times> 'a dest \<times> 'b content"

text{*  
  A packet therefore has two parameters, the first being the address,
  the second the content. These must of course be specified before we
  can generate concrete test data later.
*}

text{* 
  Alternatively, a packet could have been modelled as a record
  type. But as \testgen has yet some problems handling them, we used
  the simpler tuple definition.

  In order to access the different parts of a packet directly, we
  define a couple of projectors.
*}
consts 
  id         :: "('a,'b) packet \<Rightarrow> id"
  protocol  :: "('a,'b) packet \<Rightarrow> protocol"
  src        :: "('a,'b) packet \<Rightarrow> 'a src"
  dest       :: "('a,'b) packet \<Rightarrow> 'a dest"
  data       :: "('a,'b) packet \<Rightarrow> 'b content"
 
defs 
  id_def:         "id           \<equiv> fst"
  protocol_def:  "protocol    \<equiv> fst o snd"
  src_def:        "src          \<equiv> fst o snd o snd"
  dest_def:       "dest         \<equiv> fst o snd o snd o snd"
  data_def:       "data        \<equiv> snd o snd o snd o snd"


section{* Nets *}

text{* 
  A net is modelled as an axiomatic type class, which is roughly
  speaking an axiomatic specification of a class of types (see
  \cite{nipkow.ea:isabelle:2002} for further information). Using type
  classes gives us the possibility to enable different representations
  of an address.
*}


axclass net < type  
text{* 
  We already provide some instances to the class net: function
  application, pairing, and the integers.  
*}


instance "fun" :: (net,net)net ..
instance "*" :: (net,net)net ..
instance "int" :: net ..

(*
text{* There's an important subclass of networks: those who also model addresses. This is the class of all network representations which are defined as a pair of an adress and a port. *}

axclass net_ports < net
instance fun :: (net_ports,net_ports)net ..
instance "*" :: (net_ports,net_ports)net ..
instance int :: net_ports ..
*)

text{* 
  An address is usually part of a subnetwork (e.g. the Intranet in our
  example). We model a subnet as a set of sets of addresses to reflect
  the fact that very often a subnet consists of several subnets again.
*}

types 'a subnet = "'a::net set set"

axclass port < type

instance int ::port ..

consts src_port :: "('a,'b) packet \<Rightarrow> ('c::port)"
consts dest_port :: "('a,'b) packet \<Rightarrow> 'c::port"


text{* 
  The relation {in\_subnet} (@{text "\<sqsubset>"}) checks if an address is in a
  specific subnet. It models a kind of subset relation and is defined
  as an infix operator. 
*}
constdefs
  in_subnet :: "'a adr \<Rightarrow> 'a subnet \<Rightarrow> bool"  (infixl "\<sqsubset>" 100) 
  "in_subnet a S  \<equiv> \<exists> s \<in> S. (a \<in> s)"

end
