(*****************************************************************************
 * HOL-TestGen --- theorem-prover based test case generation
 *                 http://www.brucker.ch/projects/hol-testgen/
 *                                                                            
 * Policy_Combinators.thy --- policy combinators
 * This file is part of HOL-TestGen.
 *
 * Copyright (c) 2005-2007, ETH Zurich, Switzerland
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)
(* $Id: PolicyCombinators.thy 7205 2008-01-29 19:04:00Z lukasbru $ *)

theory PolicyCombinators
imports 
PolicyCore
begin

text{* 
  In order to ease the specification of a concrete policy, we define
  some combinators. Their definition is pretty straighforward, which
  is why to not present them all here. They can be found in the
  appendix \ref{combinators}.  With these combinators, the
  specification of a policy gets very easy, and similar to tools like
  IPTables~\cite{iptables}. 
*}

subsection{* Combinators which allow or deny everything *}
text{*
  Allow every packet: 
*}
constdefs 
  allow_all    :: "('a, 'b) Rule" 
  "allow_all p  \<equiv> Some (accept p)" 

text{*
  Deny every packet:
*}
constdefs 
  deny_all :: "('a,'b) Rule"
  "deny_all p   \<equiv> Some (deny)" 
             

subsection{* 
  Combinators which allow or deny packets based on their source and/or
  destination address. 
*}
text{*
  Allow every packet coming from a specific subnet 
*}
constdefs 
  allow_all_from :: "'a::net subnet \<Rightarrow> ('a,'b) Rule"
  "allow_all_from src_net \<equiv> allow_all |` {pa. src pa \<sqsubset> src_net}"

text{*
  Deny every packet coming from a specific subnet
*}
constdefs 
  deny_all_from    :: "'a:: net subnet \<Rightarrow> ('a,'b) Rule"
  "deny_all_from src_net   \<equiv> deny_all |` {pa. src pa \<sqsubset>  src_net}"
        
text{*
  Allow every packet going to a specific subnet 
*}
constdefs
  allow_all_to    :: "'a::net subnet \<Rightarrow> ('a,'b) Rule"
  "allow_all_to dest_net   \<equiv> allow_all |` {pa. dest pa \<sqsubset>  dest_net}"

text{*
  Deny every packet going to a specific subnet
*}
constdefs 
  deny_all_to :: "'a::net subnet \<Rightarrow> ('a,'b) Rule"
  "deny_all_to dest_net \<equiv> deny_all |` {pa. dest pa \<sqsubset> dest_net}"

text{*
  Allow all packets coming from and going to specific subnets
*}
constdefs 
  allow_all_from_to    :: "'a::net subnet \<Rightarrow> 'a::net subnet \<Rightarrow> ('a,'b) Rule"
  "allow_all_from_to src_net dest_net \<equiv> allow_all |` {pa. src pa \<sqsubset> src_net \<and> dest pa \<sqsubset> dest_net}"

text{*
  Deny all packets coming from and going to specific subnets
*}
constdefs 
  deny_all_from_to    :: "'a::net subnet \<Rightarrow> 'a::net subnet \<Rightarrow> ('a,'b) Rule"
  "deny_all_from_to src_net dest_net \<equiv> deny_all |` {pa. src pa \<sqsubset> src_net \<and> dest pa \<sqsubset> dest_net}"


subsection{*
  Combinators which allow or deny packets based additionally on their
  protocol 
*}

text{*
  Allow all packets of a specific protocol 
*}
constdefs 
  allow_protocol     :: "protocol  \<Rightarrow> ('a,'b) Rule"
  "allow_protocol prot \<equiv> allow_all |` {pa. protocol pa = prot}"

text{*
  Deny all packets of a specific protocol
*}
constdefs 
  deny_protocol :: "protocol \<Rightarrow> ('a,'b) Rule"
  "deny_protocol prot \<equiv> deny_all |` {pa. protocol pa = prot}"

text{*
  Allow all packets of a specific protocol coming from a specific subnet 
*}
constdefs 
  allow_prot_from   :: "protocol \<Rightarrow> 'a::net subnet \<Rightarrow> ('a,'b) Rule"
  "allow_prot_from prot src_net  \<equiv>  allow_all |` {pa. src pa \<sqsubset>  src_net \<and> protocol pa = prot}"

text{*
  Deny all packets of a specific protocol coming from a specific subnet
*}
constdefs 
  deny_prot_from   :: "protocol \<Rightarrow> 'a::net subnet \<Rightarrow> ('a,'b) Rule"
  "deny_prot_from prot src_net  \<equiv>  deny_all |` {pa. src pa \<sqsubset>  src_net \<and> protocol pa = prot}"

text{*
  Allow all packets of a specific protocol going to a specific subnet 
*}
constdefs 
  allow_prot_to     :: "protocol \<Rightarrow> 'a::net subnet \<Rightarrow> ('a,'b) Rule"
  "allow_prot_to prot dest_net  \<equiv>  allow_all |` {pa. dest pa \<sqsubset>  dest_net \<and> protocol pa = prot}"

text{*
  Deny all packets of a specific protocol going to a specific subnet
*}
constdefs deny_prot_to :: "protocol \<Rightarrow> 'a::net subnet \<Rightarrow> ('a,'b) Rule"
              "deny_prot_to prot dest_net \<equiv> deny_all |` {pa. dest pa \<sqsubset>  dest_net \<and> protocol pa = prot}"
text{*
  Allow all packets of a specific protocol coming from and going to
  specific subnets 
*}
constdefs 
  allow_prot_from_to :: "protocol \<Rightarrow> 'a::net subnet \<Rightarrow> 'a::net subnet \<Rightarrow> ('a,'b) Rule" 
  "allow_prot_from_to prot src_net dest_net \<equiv> allow_all |` 
            {pa. src pa \<sqsubset>  src_net \<and> dest pa \<sqsubset>  dest_net \<and> protocol pa = prot}"

text{*
  Deny all packets of a specific protocol coming from and going to 
  specific subnets
*}
constdefs 
  deny_prot_from_to :: "protocol \<Rightarrow> 'a::net subnet \<Rightarrow> 'a::net subnet  \<Rightarrow> ('a,'b) Rule"
  "deny_prot_from_to prot src_net dest_net  \<equiv> deny_all|`
           {pa. src pa \<sqsubset>  src_net \<and> dest pa \<sqsubset>  dest_net \<and> protocol pa = prot}"

subsection{*
  Combinators which allow or deny everything except a protocol
*}
text{*
  Allow everything except a specific protocol:
*}
constdefs 
  allow_all_but_prot :: "protocol \<Rightarrow> ('a,'b) Rule"
  "allow_all_but_prot prot \<equiv> allow_protocol prot |` {pa. protocol pa \<noteq> prot} ++ 
                                 deny_protocol prot |`{pa. protocol pa = prot}"
text{*
  Deny everything except a specific protocol:
*}
constdefs
  deny_all_but_prot :: "protocol \<Rightarrow> ('a,'b) Rule"
  "deny_all_but_prot prot \<equiv> deny_protocol prot |` {pa. protocol pa \<noteq> prot} ++ 
                                allow_protocol prot |` {pa. protocol pa = prot}"
text{*
  Allow every packet coming from a specific subnet, except those of a
  specific protocol:
*}
constdefs 
  allow_all_but_prot_from :: "protocol \<Rightarrow> 'a::net subnet \<Rightarrow> ('a,'b) Rule"
  "allow_all_but_prot_from prot src_net \<equiv> allow_all_from src_net |` {pa. protocol pa \<noteq> prot} ++
                                              deny_all_from src_net |` {pa. protocol pa = prot}"
text{*
  Deny every packet coming from a specific subnet, except those of a
  specific protocol:
*}
constdefs 
  deny_all_but_prot_from :: "protocol \<Rightarrow> 'a::net subnet \<Rightarrow> ('a,'b) Rule"
  "deny_all_but_prot_from prot src_net \<equiv> deny_all_from src_net |` {pa. protocol pa \<noteq> prot} ++
                                           allow_all_from src_net |` {pa. protocol pa = prot}"
text{*
  Allow every packet going to a specific subnet, except those of a
  specific protocol:
*}
constdefs 
  allow_all_but_prot_to :: "protocol \<Rightarrow> 'a::net subnet \<Rightarrow> ('a,'b) Rule"
  "allow_all_but_prot_to prot dest_net \<equiv> allow_all_to dest_net |` {pa. protocol pa \<noteq> prot} ++
                                             deny_all_to dest_net |` {pa. protocol pa = prot}"
text{*
  Deny every packet going to a specific subnet, except those of a
  specific protocol:
*}
constdefs 
  deny_all_but_prot_to :: "protocol \<Rightarrow> 'a::net subnet \<Rightarrow> ('a,'b) Rule"
  "deny_all_but_prot_to prot dest_net \<equiv> 
          deny_all_to dest_net |` {pa. protocol pa \<noteq> prot} ++
          allow_all_to dest_net |` {pa. protocol pa = prot}"

text{*
  Allow every packet coming from and going to a specific subnet,
  except those of a specific protocol:
*}
constdefs 
  allow_all_but_prot_from_to :: "protocol \<Rightarrow> 'a::net subnet \<Rightarrow> 'a::net subnet \<Rightarrow> ('a,'b) Rule"
  "allow_all_but_prot_from_to prot src_net dest_net \<equiv> 
           allow_all_from_to src_net dest_net |` {pa. protocol pa \<noteq> prot} ++
           deny_all_from_to src_net dest_net |` {pa. protocol pa = prot}"

text{*
  Deny every packet coming from and going to a specific subnet, except
  those of a specific protocol:
*}
constdefs 
  deny_all_but_prot_from_to :: "'a::net subnet \<Rightarrow> 'a::net subnet \<Rightarrow> protocol \<Rightarrow> ('a,'b) Rule"
  "deny_all_but_prot_from_to src_net dest_net prot \<equiv> 
           deny_all_from_to src_net dest_net |` {pa. protocol pa \<noteq> prot} ++
           allow_all_from_to src_net dest_net |` {pa. protocol pa = prot}"


subsection{* 
  Combinators which allow or deny everything of a set of protocols
*}
text{*
  Allow all packets of a specific set of protocols coming from and
  going to specific subnets 
*}
constdefs 
  allow_prots_from_to :: "protocol set \<Rightarrow> 'a::net subnet \<Rightarrow> 'a::net subnet \<Rightarrow>  ('a,'b) Rule" 
  "allow_prots_from_to prots src_net dest_net \<equiv> allow_all |` 
            {pa. src pa \<sqsubset>  src_net \<and> dest pa \<sqsubset>  dest_net \<and> protocol pa \<in> prots}"
text{*
  Deny all packets of a specific set of protocols coming from and
  going to specific subnets
*}
constdefs 
  deny_prots_from_to :: "protocol set \<Rightarrow> 'a::net subnet \<Rightarrow> 'a::net subnet \<Rightarrow> ('a,'b) Rule"
  "deny_prots_from_to prots src_net dest_net \<equiv> deny_all |` 
           {pa. src pa \<sqsubset>  src_net \<and> dest pa \<sqsubset>  dest_net \<and> protocol pa \<in> prots}"

subsection{* 
  Combinators which allow or deny everything except of a set of protocols
*}
text {*
  Allow every packet coming from and going to a specific subnet,
  except those of a specific set of protocols
  *}
constdefs 
  allow_all_but_prots_from_to :: " protocol set \<Rightarrow> 'a::net subnet \<Rightarrow> 'a::net subnet \<Rightarrow> ('a,'b) Rule"
  "allow_all_but_prots_from_to prots src_net dest_net \<equiv> 
           allow_all_from_to src_net dest_net |` {pa. protocol pa \<notin> prots} ++ 
           deny_all_from_to src_net dest_net |` {pa. protocol pa \<in> prots}"

text {*
  Deny every packet coming from and going to a specific subnet, except
  those of a specific set of protocols
  *}
constdefs 
  deny_all_but_prots_from_to :: "protocol set \<Rightarrow> 'a::net subnet \<Rightarrow> 'a::net subnet \<Rightarrow> ('a,'b) Rule"
  "deny_all_but_prots_from_to prots src_net dest_net \<equiv> 
            deny_all_from_to src_net dest_net |` {pa. protocol pa \<notin> prots} ++ 
            allow_all_from_to src_net dest_net |` {pa. protocol pa \<in> prots}"



text{*
  All these combinators and the default rules are put into one single
  lemma called @{text stateless_rules} to make life easier when we
  need to unfold a policy consisting of several rules. 
*}

lemmas stateless_rules =
  allow_all_def deny_all_def allow_all_from_def deny_all_from_def
  allow_all_to_def deny_all_to_def allow_all_from_to_def deny_all_from_to_def
  allow_protocol_def deny_protocol_def allow_prot_from_def deny_prot_from_def
  allow_prot_to_def deny_prot_to_def allow_prot_from_to_def deny_prot_from_to_def
  allow_all_but_prot_def deny_all_but_prot_def allow_all_but_prot_from_def
  deny_all_but_prot_from_def allow_all_but_prot_to_def deny_all_but_prot_to_def
  allow_all_but_prot_from_to_def deny_all_but_prot_from_to_def
  allow_prots_from_to_def deny_prots_from_to_def 
  allow_all_but_prots_from_to_def deny_all_but_prots_from_to_def
  map_add_def restrict_map_def

end
