(*****************************************************************************
 * HOL-TestGen --- theorem-prover based test case generation
 *                 http://www.brucker.ch/projects/hol-testgen/
 *                                                                            
 * policy.thy --- modelling policies
 * This file is part of HOL-TestGen.
 *
 * Copyright (c) 2005-2007, ETH Zurich, Switzerland
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)
(* $Id: PolicyCore.thy 7172 2008-01-24 09:49:22Z brucker $ *)

header{*Policy*}
theory PolicyCore
imports NetworkCore
begin

section{* Definitions *}

text{* 
  Having defined what a packet and a net is, the concept of a policy
  needs to be modelled. A policy is seen as a ruleset as in Table
  \ref{policy}, which specifies for each packet an action, namely if
  it should get accepted or denied by the firewall. We model these two
  possibilities by a datatype @{text "out"}. Alternatively we could
  make a further distinction between a drop and a reject if desired.
*}
datatype 'a out = accept 'a 
                | deny

text{* 
  A policy is seen as a partial mapping from packet to packet
  out. Partial mappings are modelled with the datatype @{text
  "option"}. This datatype consists of two elements, being either
  @{text "None"} or @{text "Some 'a"}.  The packets @{text p} that a
  firewall should accept, map to @{text "Some accept p"}, those that
  it should reject to @{text "Some deny"}\footnote{As the packet is
  dropped here, we don't need it anymore.}, and those packets for
  which no rule applies map to @{text "None"}. This definition enables
  us to detect packets for which no action was specified in the
  policy. For partial mappings, Isabelle provides the operator
  ``@{text "\<rightharpoonup>"}'' which is shorthand for a function from @{text "'a"}
  to @{text "'b option"}.
*}
types ('a, 'b) Policy = "('a, 'b) packet \<rightharpoonup> (('a, 'b) packet) out"   (* \<rightharpoonup> is from Map.thy *)

text{*
  A rule has the same type as a policy and usually a policy consists
  of several rules. 
*}
types ('a, 'b) Rule   = "('a, 'b) Policy"

text{*
  When combining several rules, the firewall is supposed to apply the
  first matching one. In our setting this means the first rule which
  maps the packet in question to @{text "Some packet out"}. This is
  exactly what happens when using the map-add operator (@{text "rule1
  ++ rule2"}). The only difference is that the rules must be given in
  reverse order. 
*}



text{*   
  The constant @{text p_accept} is @{text "True"} if the policy
  accepts the packet and @{text "False"} otherwise. 
*}

constdefs 
  p_accept :: "('a, 'b) packet \<Rightarrow> ('a, 'b) Policy \<Rightarrow> bool"
  "p_accept p policy \<equiv>  policy p = Some (accept p)"

end
