(*****************************************************************************
 * HOL-TestGen --- theorem-prover based test case generation
 *                 http://www.brucker.ch/projects/hol-testgen/
 *                                                                            
 * IPv4_Combinators.thy  --- modelling ipv4 addresses 
 * This file is part of HOL-TestGen.
 *
 * Copyright (c) 2005-2007, ETH Zurich, Switzerland
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)
(* $Id: IPv4_Combinators.thy 6651 2007-07-02 05:17:53Z brucker $ *)

theory DatatypeTwoNetsCombinators
  imports "../PolicyCombinators" DatatypeTwoNets
begin


text{*
  In this theory, we defines the combinators which can be used to
  define policies for stateless firewalls based on DatatypeTwoNets networks. The
  rules are generally the ones from stateless.thy, but also take port
  numbers into account. If we don't want to use the port numbers in
  our specification, we can continue to use the original rules.
*}

subsection{*
  Combinators which allow or deny packets based on their source and \/
  or destination address.
*}
text{*
  Allow every packet coming from a specific subnet on a specific port
*}
constdefs
  allow_all_from_port :: "DatatypeTwoNets subnet \<Rightarrow> port \<Rightarrow> (DatatypeTwoNets, 'b) Rule"
  "allow_all_from_port src_net s_port \<equiv> allow_all_from src_net |` {pa. src_port pa = s_port}"

text{*
  Deny every packet coming from a specific subnet on specific port
*}
constdefs 
  deny_all_from_port    :: "DatatypeTwoNets subnet \<Rightarrow> port \<Rightarrow> (DatatypeTwoNets, 'b) Rule"
  "deny_all_from_port src_net s_port  \<equiv> deny_all_from src_net |` {pa. src_port pa = s_port}"

text{*
  Allow every packet going to a specific subnet on a specific port 
*}
constdefs
  allow_all_to_port    :: "DatatypeTwoNets subnet \<Rightarrow> port \<Rightarrow> (DatatypeTwoNets, 'b) Rule"
  "allow_all_to_port dest_net d_port  \<equiv> allow_all_to dest_net |` {pa. dest_port pa = d_port}"

text{*
  Deny every packet going to a specific subnet on a specific port
*}
constdefs 
  deny_all_to_port :: "DatatypeTwoNets subnet \<Rightarrow>  port \<Rightarrow> (DatatypeTwoNets, 'b) Rule"
  "deny_all_to_port dest_net d_port \<equiv> deny_all_to dest_net|` {pa. dest_port pa = d_port}"

text{*
  Allow all packets coming from a specific port on a specific subnet
  to a specific subnet
*}
constdefs
  allow_all_from_port_to    :: "DatatypeTwoNets subnet \<Rightarrow> port \<Rightarrow> DatatypeTwoNets subnet \<Rightarrow> (DatatypeTwoNets, 'b) Rule"
  "allow_all_from_port_to src_net s_port dest_net   
       \<equiv> allow_all_from_to src_net dest_net |` {pa. src_port pa = s_port}"

text{*
  Deny every packet coming from a specific port on a specific subnet
  to a specific subnet
*}
constdefs 
  deny_all_from_port_to :: "DatatypeTwoNets subnet \<Rightarrow>  port \<Rightarrow> DatatypeTwoNets subnet \<Rightarrow> (DatatypeTwoNets, 'b) Rule"
  "deny_all_from_port_to src_net s_port dest_net  
      \<equiv> deny_all_from_to src_net dest_net|` {pa. src_port pa = s_port}"

text{*
  Allow all packets coming from a specific port on a specific subnet
  to a port on a specific subnet
*}

constdefs
  allow_all_from_port_to_port    :: "DatatypeTwoNets subnet \<Rightarrow> port \<Rightarrow> DatatypeTwoNets subnet \<Rightarrow> port \<Rightarrow> (DatatypeTwoNets, 'b) Rule"
  "allow_all_from_port_to_port src_net s_port dest_net d_port \<equiv> 
      allow_all_from_port_to src_net s_port dest_net |` {pa. dest_port pa = d_port}"

text{*
  Deny every packet coming from a specific port on a specific subnet
  to a port on a specific subnet 
*}
constdefs 
  deny_all_from_port_to_port :: "DatatypeTwoNets subnet \<Rightarrow>  port \<Rightarrow> DatatypeTwoNets subnet \<Rightarrow> port \<Rightarrow> (DatatypeTwoNets, 'b) Rule"
  "deny_all_from_port_to_port src_net s_port dest_net d_port \<equiv> 
      deny_all_from_port_to src_net s_port dest_net|` {pa. dest_port pa = d_port}"

  

text{*
  Allow all packets coming from and going to specific subnets on
  specific ports
*}
constdefs 
  allow_all_from_to_port    :: "DatatypeTwoNets subnet \<Rightarrow> port \<Rightarrow> DatatypeTwoNets subnet \<Rightarrow> port \<Rightarrow> (DatatypeTwoNets, 'b) Rule"
  "allow_all_from_to_port src_net s_port dest_net d_port \<equiv> allow_all_from_to src_net dest_net|` 
                                      {pa. src_port pa = s_port \<and> dest_port pa = d_port}"
text{*
  Deny all packets coming from and going to specific subnets on
  specific ports 
*}

constdefs 
  deny_all_from_to_port    :: "DatatypeTwoNets subnet \<Rightarrow> port \<Rightarrow> DatatypeTwoNets subnet \<Rightarrow> port \<Rightarrow>  (DatatypeTwoNets, 'b) Rule"
  "deny_all_from_to_port src_net s_port dest_net d_port \<equiv> deny_all_from_to src_net dest_net |` 
                                       {pa. src_port pa = s_port \<and> dest_port pa = d_port}"


subsection{*
  Combinators which allow or deny packets based additionally on their
  protocol 
*}

text{*
  Allow all packets of a specific protocol coming from a specific
  subnet on a specific port 
*}
constdefs 
  allow_prot_from_port   :: "protocol \<Rightarrow> DatatypeTwoNets subnet \<Rightarrow> port \<Rightarrow> (DatatypeTwoNets, 'b) Rule"
  "allow_prot_from_port prot src_net s_port  \<equiv>  allow_prot_from prot src_net |` 
                             {pa. src_port pa = s_port}"
text{*
  Deny all packets of a specific protocol coming from a specific
  subnet on a specific port
*}
constdefs 
  deny_prot_from_port   :: "protocol \<Rightarrow> DatatypeTwoNets subnet \<Rightarrow> port \<Rightarrow> (DatatypeTwoNets, 'b) Rule"
  "deny_prot_from_port prot src_net s_port \<equiv>  deny_prot_from prot src_net |` 
                             {pa. src_port pa = s_port}"
text{*
  Allow all packets of a specific protocol going to a specific subnet
  on a specific port 
*}
constdefs 
  allow_prot_to_port    :: "protocol \<Rightarrow> DatatypeTwoNets subnet \<Rightarrow> port \<Rightarrow> (DatatypeTwoNets, 'b) Rule"
  "allow_prot_to_port prot dest_net d_port  \<equiv>  allow_prot_to prot dest_net |` 
                             {pa. dest_port pa = d_port}"
text{*
  Deny all packets of a specific protocol going to a specific subnet
  on a specific port
*}
constdefs 
  deny_prot_to_port :: "protocol \<Rightarrow> DatatypeTwoNets subnet \<Rightarrow> port \<Rightarrow> (DatatypeTwoNets, 'b) Rule"
  "deny_prot_to_port prot dest_net d_port \<equiv> deny_prot_to prot dest_net |` 
                             {pa. dest_port pa = d_port}"

text{*
  Allow all packets of a specific protocol coming from a specific
  subnet on a specific port and going to a specific subnet
*}
constdefs 
  allow_prot_from_port_to :: "protocol \<Rightarrow> DatatypeTwoNets subnet \<Rightarrow> port \<Rightarrow> DatatypeTwoNets subnet \<Rightarrow> (DatatypeTwoNets, 'b) Rule" 
  "allow_prot_from_port_to pr src_net s_port dest_net \<equiv> allow_prot_from_to pr src_net dest_net |` 
                                                        {pa. src_port pa = s_port}"
text{*
  Deny all packets of a specific protocol coming from a specific
  subnet on a specific port and going to a specific subnet
*}
constdefs 
  deny_prot_from_port_to :: "protocol \<Rightarrow> DatatypeTwoNets subnet \<Rightarrow> port \<Rightarrow> DatatypeTwoNets subnet  \<Rightarrow> (DatatypeTwoNets, 'b) Rule"
  "deny_prot_from_port_to pr src_net s_port dest_net \<equiv> deny_prot_from_to pr src_net dest_net |`
                                             {pa. src_port pa = s_port}"
text{*
  Allow all packets of a specific protocol coming from a specific
  subnet and going to a specific subnet on a specific port
*}
constdefs 
  allow_prot_from_to_port :: "protocol \<Rightarrow> DatatypeTwoNets subnet \<Rightarrow> DatatypeTwoNets subnet \<Rightarrow> port \<Rightarrow> (DatatypeTwoNets, 'b) Rule" 
  "allow_prot_from_to_port pr src_net dest_net d_port \<equiv> allow_prot_from_to pr src_net dest_net |` 
                                                        {pa. dest_port pa = d_port}"
text{*
  Deny all packets of a specific protocol coming from a specific
  subnet and going to a specific subnet on a specific port
*}
constdefs 
  deny_prot_from_to_port :: "protocol \<Rightarrow> DatatypeTwoNets subnet \<Rightarrow> DatatypeTwoNets subnet \<Rightarrow> port  \<Rightarrow> (DatatypeTwoNets, 'b) Rule"
  "deny_prot_from_to_port pr src_net dest_net d_port \<equiv> deny_prot_from_to pr src_net dest_net |`
                                             {pa. dest_port pa = d_port}"

text{*
  Allow all packets of a specific protocol coming from a specific
  subnet on a specific port and going to a specific subnet on a
  specific port
*}
constdefs 
  allow_prot_from_port_to_port :: "protocol \<Rightarrow> DatatypeTwoNets subnet \<Rightarrow> port \<Rightarrow> DatatypeTwoNets subnet \<Rightarrow> port \<Rightarrow> (DatatypeTwoNets, 'b) Rule" 
  "allow_prot_from_port_to_port pr src_net s_port dest_net d_port \<equiv> 
                          allow_all_from_port_to_port src_net s_port dest_net d_port|` 
                                                        {pa. protocol pa = pr}"


text{*
  Deny all packets of a specific protocol coming from a specific
  subnet and going to a specific subnet on a specific port
*}
constdefs 
  deny_prot_from_port_to_port :: "protocol \<Rightarrow> DatatypeTwoNets subnet \<Rightarrow> port \<Rightarrow> DatatypeTwoNets subnet \<Rightarrow> port  \<Rightarrow> (DatatypeTwoNets, 'b) Rule"
  "deny_prot_from_port_to_port pr src_net s_port dest_net d_port \<equiv> 
                  deny_all_from_port_to_port src_net s_port dest_net d_port|`
                                                                   {pa. protocol pa = pr}"

constdefs 
  allow_prots_from_to_port :: "DatatypeTwoNets subnet \<Rightarrow> DatatypeTwoNets subnet \<Rightarrow> (protocol \<times> port) set \<Rightarrow> (DatatypeTwoNets ,'b) Rule" 
  "allow_prots_from_to_port src_net dest_net protsports \<equiv> allow_all |` 
            {pa. src pa \<sqsubset>  src_net \<and> dest pa \<sqsubset>  dest_net \<and> (\<exists> pp\<in>protsports. protocol pa = fst pp \<and> dest_port pa = snd pp)}"

constdefs 
  allow_prots_from_port_to :: "DatatypeTwoNets subnet \<Rightarrow> DatatypeTwoNets subnet \<Rightarrow> (protocol \<times> port) set \<Rightarrow> (DatatypeTwoNets,'b) Rule" 
  "allow_prots_from_port_to src_net dest_net protsports \<equiv> allow_all |` 
            {pa. src pa \<sqsubset>  src_net \<and> dest pa \<sqsubset>  dest_net \<and> (\<exists> pp\<in>protsports. protocol pa = fst pp \<and> src_port pa = snd pp)}"




text{* 
  As before, we put all the rules into one lemma called @{text
  "DatatypeTwoNets_rules"} to ease writing later. 
*}
lemmas DatatypeTwoNetsCombinators = 
  allow_all_from_port_def deny_all_from_port_def 
  allow_all_to_port_def deny_all_to_port_def allow_all_from_to_port_def
  deny_all_from_to_port_def allow_prot_from_port_def deny_prot_from_port_def
  allow_prot_to_port_def deny_prot_to_port_def allow_prot_from_port_to_def
  deny_prot_from_port_to_def allow_prot_from_to_port_def deny_prot_from_to_port_def
  allow_prots_from_to_port_def allow_prots_from_port_to_def

end
