(*****************************************************************************
 * HOL-TestGen --- theorem-prover based test case generation
 *                 http://www.brucker.ch/projects/hol-testgen/
 *                                                                            
 * FTP.thy --- modelling the ftp protocol
 * This file is part of HOL-TestGen.
 *
 * Copyright (c) 2005-2007, ETH Zurich, Switzerland
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)
(* $Id: FTP.thy 7205 2008-01-29 19:04:00Z lukasbru $ *)

header {* FTP *}
theory FTP
imports 
  "Stateful" 
  "IPv4"
  "PortCombinators"
begin

text{*
  The File Transfer Protocol FTP~\cite{rfc_ftp} is a well known
  example of a protocol which uses dynamic ports and is therefore a
  natural choice to use as an example for our model.
  \begin{figure} 
    \centering
    \includegraphics[scale=0.28]{ftp_AB} 
    \caption {The modelled FTP-Protocol}
    \label{ftp_AB}
  \end{figure}

  We model only a simplified version of the FTP protocol over IPv4
  networks, still containing all messages that matter for our
  purposes. It consists of the following four messages (see Figure
  \ref{ftp_AB}):

  \begin{enumerate} 
    \item @{text "init"}: The client contacts the server indicating
          his wish to get some data.
    \item @{text "port_request p"}: The client, usually after having
          received an acknowledgement of the server, indicates a port
          number on which he wants to receive the data.
    \item @{text "ftp_data"}: The server sends the requested data over
          the new channel. There might be an arbitrary number of such
          messages, including zero.
      \item @{text "close"}: The client closes the connection. The
          dynamic port gets closed again.
  \end{enumerate}

  The content field of a packet therefore now consists of these four
  messages plus a default one.
*}
datatype ftp_msg = init 
                         | port_request port 
                         | ftp_data 
                         | close 
                         | other


text{* 
  We now also make use of the ID field of a packet. It is used as
  session ID and we make the assumption that they are all unique. 

  At first, we need some predicates which check if a packet is a
  specific FTP message and has the correct session ID. 
*}
constdefs
  is_init :: "id \<Rightarrow> ('a, ftp_msg ) packet \<Rightarrow> bool"
  "is_init i p \<equiv>  id p = i \<and> data p = init"

constdefs
  is_port_request :: "id  \<Rightarrow> port \<Rightarrow>('a, ftp_msg) packet \<Rightarrow> bool"
  "is_port_request i port p \<equiv>  id p = i \<and> data p = port_request port"

constdefs
  is_data :: "id \<Rightarrow> ('a, ftp_msg) packet \<Rightarrow> bool"
  "is_data i p \<equiv>  id p = i \<and> data p = ftp_data"

constdefs
  is_close :: "id \<Rightarrow> ('a, ftp_msg) packet \<Rightarrow> bool"
  "is_close i p \<equiv> id p = i \<and> data p = close"



text{*
  We now have to model the respective state transitions. It is
  important to note that state transitions themselves allow all
  packets which are allowed by the policy, not only those which are
  allowed by the protocol. Their only task is to change the policy. As
  an alternative, we could have decided that they only allow packets
  which follow the protocol (e.g. come on the correct ports), but this
  should in our view rather be reflected in the policy itself.

  Of course, not every message changes the policy. In such cases, we
  do not have to model different state transitions, one is enough. In
  this example, only messages 2 and 4 need special transitions. The
  default says that if the policy accepts the packet, it is added to
  the history, otherwise it is simply dropped. The policy remains the
  same in both cases. 
*}
consts FTP_ST_F :: "((ipv4,ftp_msg) history ,ipv4, ftp_msg) FWStateTransition"
recdef FTP_ST_F "{}"
  FTP_ST_F_1: "FTP_ST_F ((a,ftp,c,d,e), (InL, policy)) = 
             (if p_accept (a,ftp,c,d,e) policy then 
                   Some ((a,ftp,c,d,e)#InL,policy) 
              else 
                   Some (InL,policy))"

  FTP_ST_F_2: "FTP_ST_F (x, (InL,policy)) = None"


text{* 
  The second message is the port request. If the packet is allowed by
  the policy, and if and only if there is an opened but not yet closed
  FTP-Session with the same session ID, we change our policy such that
  the requested port is opened. If our policy allows the packet but
  there is no open protocol run, we do allow the packet but do not
  open the requested port. 
*}
consts FTP_ST_2 :: "((ipv4,ftp_msg) history, ipv4, ftp_msg) FWStateTransition"
recdef FTP_ST_2 "{}"
  FTP_ST_2_1: "FTP_ST_2 ((a,ftp,c,d,port_request port_r), (InL,policy )) = 
                   (if p_accept (a,ftp,c,d,port_request port_r) policy then
                       (if not_before  (is_close a) (is_init a) InL then 
                             Some ((a,ftp,c,d,port_request port_r)#InL, policy ++ 
                                  (allow_prot_from_to_port ftp (subnet_of d) (subnet_of c) port_r))
                        else Some ((a,ftp,c,d,port_request port_r)#InL,policy))
                    else Some (InL,policy))" 

FTP_ST_2_2: "FTP_ST_2 (x, (InL,policy)) = None"

text{* 
  In the last message, we need to close a port which we do not know
  directly. It has only been specified in a preceding port\_request
  message. Therefore a predicate is needed which checks if there is an
  open protocol run with an opened port 
*}
constdefs 
  port_open :: "(ipv4, ftp_msg) history \<Rightarrow> id \<Rightarrow> port \<Rightarrow> bool"
  "port_open L a p \<equiv> not_before (is_close a) (is_port_request a p) L"


text{* 
  This transition is the trickiest one. We need to close the port wich
  has been opened but not yet closed by a packet with the same session
  ID. Here we use the assumption that they are supposed to be
  unique. 
*}

consts FTP_ST_4 :: "((ipv4, ftp_msg) history, ipv4, ftp_msg) FWStateTransition"
recdef FTP_ST_4 "{}"
  FTP_ST_4_1: "FTP_ST_4 ((a,ftp,c,d,close), (InL,policy)) = 
                   (if (p_accept (a,ftp,c,d,close) policy) then 
                         (if (\<exists> p. port_open InL a p) then 
                              Some((a,ftp,c,d,close)#InL, policy ++ 
deny_prot_from_to_port ftp (subnet_of d) (subnet_of c) (Eps (\<lambda> p. port_open InL a p)))
                          else Some ((a,ftp,c,d,close)#InL, policy))
                    else Some (InL,policy))"

FTP_ST_4_2: "FTP_ST_4 (x, (InL,policy)) = None"

text{* 
  This transition introduces some kind of inconsistency. If the port
  that was requested was already open to start with, it gets closed
  here. The tester should be aware of this fact.

  This transition has also some other consequences. The Hilbert
  epsilon operator @{text "Eps"}, also written as @{text "SOME"},
  returns an arbitrary object for which the following predicate is
  @{text "True"} and is undefined otherwise. We use it to get the
  number of the port which we want to close. With the if-condition it
  is assured that such a port exists, but we might have problems if
  there are several of them. However, due to our assumption that the
  session IDs are unique, there won't be a problem as long as we do
  not open several ports in one single protocol run. This should not
  occur by the definition of the protocol, but if it does, which might
  happen if we want to test illegal protocol runs, some proof work
  might be needed.

  The full state machine of an FTP-protocol is now the @{text orelse}
  of the single transitions. 
*}
constdefs 
  FTP_ST :: "((ipv4, ftp_msg) history, ipv4, ftp_msg) FWStateTransition"
  "FTP_ST \<equiv> FTP_ST_2 orelse  
                FTP_ST_4 orelse 
                FTP_ST_F"


text{* 
  An arbitrary repetition of protocol runs can be modelled with the
  operator @{term repeat}.
*}
constdefs 
  ALL_ST :: "((ipv4,ftp_msg) history, ipv4, ftp_msg) FWStateTransition"
  "ALL_ST \<equiv> repeat FTP_ST"

text{* 
  Now we specify our test scenario. We could test
  \begin{itemize}
    \item one correct FTP-Protocol run,
    \item several runs after another,
    \item several runs interleaved,
    \item an illegal protocol run, or
    \item several illegal protocol runs.
  \end{itemize}

  We only do the the simplest case: one correct protocol run. As soon
  as we are modelling the more complicated ones, the state space grows
  rapidly and there will be problems with creating test cases within
  reasonable time.
*}

text{*
  There are four different states which are modelled as a datatype.
*}
datatype ftp_states = S0 | S1 | S2 | S3


text{* 
  The following constant is @{text "True"} for all sets which are
  correct FTP runs for a given source and destination address, ID, and
  data-port number. The empty trace is not included as it wouldn't
  make any sense to test this one. 
*}
consts
  is_ftp :: "ftp_states \<Rightarrow> 'a adr \<Rightarrow> 'a adr \<Rightarrow> id \<Rightarrow> port \<Rightarrow> ('a,ftp_msg) history \<Rightarrow> bool"
primrec
  "is_ftp H c s i p [] = (H=S3)"
  "is_ftp H c s i p (x#InL) = (\<lambda> (id,pr,sr,de,co). (((pr = ftp \<and> id = i \<and> (
        (H=S2 \<and> sr = c \<and> de = s \<and> co = init \<and> is_ftp S3 c s i p InL) \<or> 
        (H=S1 \<and> sr = c \<and> de = s \<and> co = port_request p \<and>  is_ftp S2 c s i p InL) \<or> 
        (H=S1 \<and> sr = s \<and> de = (fst c,p) \<and> co= ftp_data \<and> is_ftp S1 c s i p InL) \<or> 
        (H=S0 \<and> sr = c \<and> de = s \<and> co = close  \<and> is_ftp S1 c s i p InL) ))))) x"
text{* 
  This definition is crucial for specifying what we actually want to
  test. Extending it produces more test cases but increases the time
  necessary to create them and vice-versa. It is a good idea for the
  tester to experiment a little bit around here. 
*}


text{* 
  The following constant then returns a set of all the historys which
  denote such a normal behaviour FTP run, again for a given source and
  destination address, ID, and data-port. 
*}
constdefs 
  NB_ftp :: "'a src \<Rightarrow> 'a dest \<Rightarrow> id \<Rightarrow> port  \<Rightarrow> ('a,ftp_msg) history set"
  "NB_ftp s d i p \<equiv> {x. (is_ftp S0 s d i p x)}"
end
