(*****************************************************************************
 * HOL-TestGen --- theorem-prover based test case generation
 *                 http://www.brucker.ch/projects/hol-testgen/
 *                                                                            
 * VOIP.thy --- modelling VoIP
 * This file is part of HOL-TestGen.
 *
 * Copyright (c) 2005-2007, ETH Zurich, Switzerland
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)
(* $Id: VOIP.thy 7205 2008-01-29 19:04:00Z lukasbru $ *)


header {* VoIP *}
theory VOIP
  imports "Stateful" "IPv4" "PortCombinators"
begin

text{*

  After the FTP-Protocol which was rather simple we show the strength
  of the model with a more current and especially much more
  complicated example, namely Voice over IP (VoIP). VoIP is
  standardized by the ITU-T under the name H.323, which can be seen as
  an "umbrella standard" which aggregates standards for multimedia
  conferencing over packet-based networks (for a good overview of the
  protocol suite, see \cite{switch_basic}).  H.323 poses many problems
  to firewalls. These problems include (taken from~\cite{vtel}):
  \begin{itemize}
    \item An H.323 call is made up of many different simultaneous
          connections.
    \item Most connections are made to dynamic ports.
    \item The addresses and port numbers are exchanged within
          the data stream of the next higher connection. 
    \item Calls can be initiated from outside the firewall.
  \end{itemize}

 \begin{figure} 
   \centering
   \includegraphics[scale=0.4]{voip} 
   \caption {The modelled VoIP-Protocol}
   \label{voip}
  \end{figure}

  Again we only consider a simplified VoIP scenario with the following
  seven messages which are grouped into four subprotocols (see Figure
  \ref{voip}):
  \begin{itemize}
    \item Registration and Admission (H.225, port 1719): The caller
          contacts its gatekeeper with a call request. The gatekeeper
          either rejects or confirms the request, returning the
          address of the callee in the latter case.
  
          \begin{itemize}
            \item Admission Request (ARQ)
            \item Admission Reject (ARJ)
            \item Admission Confirm (ACF) @{text "'a"}
          \end{itemize}
    \item Call Signaling (Q.931, port 1720) The caller and the callee 
          agree on the dynamic ports over which the call will take 
          place.  
          \begin{itemize}
            \item Setup @{text "port"}
            \item Connect @{text "port"}
          \end{itemize}
    \item Stream (dynamic ports). The call itself. In reality, several 
          connections are used here.
    \item Fin (port 1720).
  \end{itemize}


  The two main differences to FTP are:
  \begin{itemize}
    \item In VoIP, we deal with three different entities: the caller, 
          the callee, and the gatekeeper.
    \item We do not know in advance which entity will close the 
          connection.
  \end{itemize}
  
  We model the protocol as seen from a firewall at the caller, namely
  we are not interested in the messages from the callee to its
  gatekeeper. Incoming calls are not modelled either, they would
  require a different set of state transitions.  
*}


text{* 
  The content of a packet now consists of one of the seven messages or
  a default one. It is parameterized with the type of the address that
  the gatekeeper returns.
*}


datatype 'a voip_msg =  ARQ
                     | ACF 'a 
                     | ARJ
                     | Setup port 
                     | Connect port 
                     | Stream 
                     | Fin 
                     | other
text{* 
  As before, we need operators which check if a packet contains a
  specific content and ID, respectively if such a packet has appeared
  in the trace. 
*}


constdefs
  is_arq :: "id \<Rightarrow> ('a, 'b voip_msg) packet \<Rightarrow> bool"
  "is_arq i p \<equiv>  id p = i \<and> data p = ARQ"

  was_arq :: "id \<Rightarrow> ('a, 'b voip_msg) history \<Rightarrow> bool"
  "was_arq i L \<equiv> \<exists> p. (p mem L \<and> is_arq i p)"

  is_fin :: "id \<Rightarrow> ('a, 'b voip_msg) packet \<Rightarrow> bool"
  "is_fin i p \<equiv> id p = i \<and> data p = Fin"

  is_connect :: "id \<Rightarrow> port \<Rightarrow> ('a, 'b voip_msg) packet \<Rightarrow> bool"
  "is_connect i port p \<equiv>  id p = i \<and> data p = Connect port"

  is_setup :: "id \<Rightarrow> port \<Rightarrow> ('a, 'b voip_msg) packet \<Rightarrow> bool"
  "is_setup i port p \<equiv>  id p = i \<and> data p = Setup port"


text{* 
  We need also an operator @{text ports_open} to get access to the two
  dynamic ports.
*}
constdefs 
  ports_open :: "id  \<Rightarrow>  port \<times> port \<Rightarrow> (ipv4, 'a voip_msg) history \<Rightarrow> bool"
  "ports_open i p L \<equiv> (not_before (is_fin i) (is_setup i (fst p)) L) \<and> 
                             not_before (is_fin i) (is_connect i (snd p)) L"


text{*
  The first state transition is for those messages which do not change
  the policy. In this scenario, this only happens for the Stream
  messages. 
*}
consts VOIP_ST_F :: "((ipv4, 'b voip_msg) history, ipv4, 'b voip_msg) FWStateTransition"
recdef VOIP_ST_F "{}"
  VOIP_ST_F_1: "VOIP_ST_F ((a,voip,c,d,e), (InL, policy)) = 
                      (if p_accept (a,voip,c,d,e) policy then 
                          Some ((a,voip,c,d,e)#InL,policy) 
                       else 
                          Some (InL,policy))"

  VOIP_ST_F_2: "VOIP_ST_F (x, (InL,policy)) = None"


text {*
    If our policy accepts the ARQ packet, we have to assure that we
    will accept the returning packet of the gatekeeper.
*}
consts VOIP_ST_1 :: "((ipv4, 'b voip_msg) history, ipv4, 'b voip_msg) FWStateTransition"
recdef VOIP_ST_1 "{}"
  VOIP_ST_1_1: "VOIP_ST_1 ((a,voip,c,d,ARQ), (InL, policy)) =
                        (if (p_accept (a,voip,c,d,ARQ) policy)
                         then Some ((a,voip,c,d, ARQ)#InL, policy ++ 
                                   allow_prot_from_to voip (subnet_of d) (subnet_of c))
                         else Some (InL,policy))"

  VOIP_ST_1_2: "VOIP_ST_1 (x, (InL,policy)) = None"



text{* 
  And if the gatekeeper answers, no matter if it's a good or bad
  answer, we can close the channel again. If the answer was positive
  (ACF), we allow the caller to contact the callee and get contacted
  by him over port 1720. 
*}
consts VOIP_ST_2 :: "((ipv4, 'b voip_msg) history, ipv4, 'b voip_msg) FWStateTransition"
recdef VOIP_ST_2 "{}"
  VOIP_ST_2_1: "VOIP_ST_2 ((a,voip,c,d,ARJ),(InL, policy)) = 
                         (if (p_accept (a,voip,c,d,ARJ) policy) 
                          then (if (not_before (is_fin a) (is_arq a) InL)
                                then Some ((a,voip,c,d,ARJ)#InL, 
                                      policy ++ deny_prot_from_to voip (subnet_of c) (subnet_of d))
                                else Some ((a,voip,c,d,ARJ)#InL,policy))
                          else Some (InL,policy))"

  VOIP_ST_2_2: "VOIP_ST_2 (x, (InL,policy)) = None"


consts VOIP_ST_3 :: "((ipv4,ipv4_ip voip_msg) history, ipv4, ipv4_ip voip_msg) 
                            FWStateTransition"
recdef VOIP_ST_3 "{}"
  VOIP_ST_3_1: "VOIP_ST_3 ((a,voip,c,d,ACF callee), (InL, policy)) =
             (if (p_accept (a,voip,c,d,ACF callee) policy)
              then Some ((a,voip,c,d,ACF callee)#InL, policy ++ 
                        allow_prot_from_to_port voip (subnet_of_ip callee) (subnet_of d) 1720 ++
                        allow_prot_from_to_port voip (subnet_of d) (subnet_of_ip callee) 1720 ++
                        deny_prot_from_to voip (subnet_of d) (subnet_of c))
              else Some (InL,policy))"

  VOIP_ST_3_2: "VOIP_ST_3 (x, (InL,policy)) = None"

text{* 
  In the Setup message, the caller specifies the port on which he
  wants the connection to take place so we need to open it for
  incoming VoIP messages. 
*}
consts VOIP_ST_4 :: "((ipv4, 'b voip_msg) history, ipv4, 'b voip_msg) FWStateTransition"
recdef VOIP_ST_4 "{}"
  VOIP_ST_4_1: "VOIP_ST_4 ((a,voip,c,d, Setup port), (InL, policy)) =
             (if (p_accept (a,voip,c,d,Setup port) policy)
              then Some ((a,voip,c,d,Setup port)#InL, policy ++ 
                        allow_prot_from_to_port voip (subnet_of d) (subnet_of c) port)
              else Some (InL,policy))"

   VOIP_ST_4_2: "VOIP_ST_4 (x, (InL,policy)) = None"

text{* 
  The same happens after the Connect message of the callee. 
*}
consts VOIP_ST_5 :: "((ipv4, 'b voip_msg) history, ipv4, 'b voip_msg) FWStateTransition"
recdef VOIP_ST_5 "{}"
  VOIP_ST_5_1: "VOIP_ST_5 ((a,voip,c,d, Connect port), (InL, policy)) =
               (if (p_accept (a,voip,c,d,Connect port) policy)
                then Some ((a,voip,c,d,Connect port)#InL, policy ++ 
                          allow_prot_from_to_port voip (subnet_of d) (subnet_of c) port)
                else Some (InL,policy))"

   VOIP_ST_5_2: "VOIP_ST_5 (x, (InL,policy)) = None"


text{* 
  As we do not know which entity closes the connection, we define an
  operator which checks if the closer is the caller.
*}
consts 
  src_is_initiator :: "id \<Rightarrow> ipv4 \<Rightarrow> (ipv4,'b voip_msg) history \<Rightarrow> bool"
primrec "src_is_initiator i a [] = False"
        "src_is_initiator i a (p#S) =  (if ((id p = i) \<and> 
                                            (\<exists> port. data p = Setup port) \<and> 
                                            ((fst (src p) = fst a))) then 
                                       True else src_is_initiator i a S)"

text{*
  In the FIN message, we have to close all the previously opened
  ports. This works as in the FTP close message, only a little bit
  more complicated. 
*}
consts VOIP_ST_7 ::" ((ipv4, 'b voip_msg) history, ipv4, 'b voip_msg) FWStateTransition"
recdef VOIP_ST_7 "{}"
  VOIP_ST_7_1: "VOIP_ST_7 ((a,voip,c,d,Fin), (InL,policy)) = 
 (if (p_accept (a,voip,c,d,Fin) policy) then
       (if \<exists> p1 p2. ports_open a (p1,p2) InL then (
           (if src_is_initiator a c InL
           then (Some ((a,voip,c,d,Fin)#InL, policy ++ 
(deny_prot_from_to_port voip (subnet_of c) (subnet_of d) 1720) ++
(deny_prot_from_to_port voip (subnet_of c) (subnet_of d) (snd (SOME p. ports_open a p InL)))++
(deny_prot_from_to_port voip (subnet_of d) (subnet_of c) (fst (SOME p. ports_open a p InL)))))
           else (Some ((a,voip,c,d,Fin)#InL, policy ++ 
(deny_prot_from_to_port voip (subnet_of c) (subnet_of d) 1720) ++
(deny_prot_from_to_port voip (subnet_of c) (subnet_of d) (fst (SOME p. ports_open a p InL))) ++
(deny_prot_from_to_port voip (subnet_of d) (subnet_of c) (snd (SOME p. ports_open a p InL)))))))
       else
           (Some ((a,voip,c,d,Fin)#InL,policy)))
  else Some (InL,policy))"

 VOIP_ST_7_2: "VOIP_ST_7 (x,(InL,policy)) = None"


text{* 
  A full VoIP protocol is the orelse of all the state transitions. 
*}
constdefs VOIP_ST :: "((ipv4, ipv4_ip voip_msg) history, ipv4, ipv4_ip voip_msg) 
                             FWStateTransition"

"VOIP_ST \<equiv> VOIP_ST_1 orelse 
                VOIP_ST_2 orelse 
                VOIP_ST_3 orelse
                VOIP_ST_4 orelse 
                VOIP_ST_5 orelse 
                VOIP_ST_7 orelse
                VOIP_ST_F"


text{* 
  For a full protocol run, six states are needed. 
*}
datatype voip_states = S0 | S1 | S2 | S3 | S4 | S5



text{*   
  The constant @{text "is_voip"} checks if a trace corresponds to a
  legal VoIP protocol, given the IP-addresses of the three entities,
  the ID, and the two dynamic ports. 
*}

consts is_voip :: "voip_states \<Rightarrow> ipv4_ip \<Rightarrow> ipv4_ip \<Rightarrow> ipv4_ip \<Rightarrow> id \<Rightarrow> port \<Rightarrow> port \<Rightarrow> 
                  (ipv4, ipv4_ip voip_msg) history \<Rightarrow> bool"
primrec 
   "is_voip H s d g i p1 p2 [] = (H = S5)"
   "is_voip H s d g i p1 p2 (x#InL) = (\<lambda> (id,pr,sr,de,co). (((pr = voip \<and> id = i \<and> (
(H = S1 \<and> sr = (s,1719) \<and> de = (g,1719) \<and> co = ARQ \<and> is_voip S5 s d g i p1 p2 InL) \<or> 
(H = S0 \<and> sr = (g,1719) \<and> de = (s,1719) \<and> co = ARJ \<and> is_voip S1 s d g i p1 p2 InL) \<or> 
(H = S2 \<and> sr = (g,1719) \<and> de = (s,1719) \<and> co = ACF d \<and> is_voip S1 s d g i p1 p2 InL) \<or>
(H = S3 \<and> sr = (s,1720) \<and> de = (d,1720) \<and> co = Setup p1 \<and> is_voip S2 s d g i p1 p2 InL) \<or> 
(H = S4 \<and> sr = (d,1720) \<and> de = (s,1720) \<and> co = Connect p2 \<and> is_voip S3 s d g i p1 p2 InL) \<or>
(H = S4 \<and> sr = (s,p1) \<and> de = (d,p2) \<and> co = Stream \<and> is_voip S4 s d g i p1 p2 InL) \<or>
(H = S4 \<and> sr = (d,p2) \<and> de = (s,p1) \<and> co = Stream \<and> is_voip S4 s d g i p1 p2 InL) \<or>
(H = S0 \<and> sr = (d,1720) \<and> de = (s,1720) \<and> co = Fin \<and> is_voip S4 s d g i p1 p2 InL) \<or> 
(H = S0 \<and> sr = (s,1720) \<and> de = (d,1720) \<and> co = Fin \<and> is_voip S4 s d g i p1 p2 InL)))))) x"
 
text{* 
  Finally, @{text "NB_voip"} returns the set of protocol traces which
  correspond to a correct protocol run given the three addresses, the
  ID, and the two dynamic ports.
*}
constdefs 
  NB_voip :: "ipv4_ip \<Rightarrow> ipv4_ip \<Rightarrow> ipv4_ip \<Rightarrow> id  \<Rightarrow> port \<Rightarrow> port \<Rightarrow> (ipv4, ipv4_ip voip_msg) history set"
  "NB_voip s d g i p1 p2\<equiv> {x. (is_voip S0 s d g i p1 p2 x)}"

end
