(*****************************************************************************
 * HOL-TestGen --- theorem-prover based test case generation
 *                 http://www.brucker.ch/projects/hol-testgen/
 *                                                                            
 * TestGen.ML  --- test data/case generation routines 
 * This file is part of HOL-TestGen.
 *
 * Copyright (c) 2005-2007 ETH Zurich, Switzerland
 *               2008 Achim D. Brucker, Germanycat 
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)
(* $Id: TestGen.ML 8103 2008-06-12 05:15:26Z brucker $ *)

(* Module Test *)

infix 4 
   add_test_case  del_test_case  add_test_data  del_test_data 

infix 0 COLLECT_SS
     
signature TESTGEN =
sig
   val trace_test         : bool ref
   (* switches on debug information *)
   val is_thyp : Term.term -> bool

   (* Core data structure for the test environment.
      Contains tables, parameters, all sorts of "screws"
      to control the test generation process. *)
   type testinfo =
               {depth   : int, (* data structure depth *)
                breadth : int, (* no of free's in testspec *)
                test_thm_tab    : thm Symtab.table, (* name to testthm *)

                bound           : int, (* global bound for data statements *)
                case_breadth    : int, (* no of test data per case,
                                          weakening uniformity *)
                iterations      : int, (* number of random attempts to solve a case *)

                type_range_bound: int,      (* effectively used elements 
                                               type grounding list *)
                type_candicates : typ list, (* type grounding list *)
 
                cod_term_tab    : (typ list -> term) Symtab.table,
                                       (* random generators for terms,
                                          tab assigns to each type constuctor 
                                          name a generator that gets the
                                          instances of type constructor as args.*)
                abstest_data_tab: (thm list) Symtab.table, 
                                       (* assigns name to list of 
                                          abstract (local) test data statements *)
                test_data_tab   : (thm list) Symtab.table};


   type testenv
  
   val  rep_testenv       : testenv -> testinfo
   val  mk_testenv        : testinfo -> testenv

   val  mt_testenv        : testenv
 
   val  add_test_case     : testenv * (string*thm)  -> testenv
   (*   add test theorem of "name".
        The test-theorem is assumed to consist of either clauses in TNF or 
        of Test Hypothesises. *)
       
   val  del_test_case     : testenv * string -> testenv
   val  add_test_data     : testenv * (string * thm)  -> testenv
   (* used to insert abstract test data registrated under
      "string" into the test environment; abstract test data
      (or "local test theorems") were used in gen_test_data *)
   val  del_test_data     : testenv * string  -> testenv
   (* erase abstract test data from test environment *)

   val  merge_testenv     : testenv * testenv -> testenv

   val  gen_test_case_tac : theory -> clasimpset -> int -> int 
                                   -> string  -> tactic
   (* "gen_test_case_tac thy clasimp depth breadth progname" is a 
      tactic that converts all subgoals of current proof state into 
      test-normal form - i.e. a conjunctive normal form
      (instead of DNF), where the conclusions have the
      form "P(progname x1 .. xn)", where progname does
      not occur in the premisses (the "constraints") 
      of the test. As a prerequisite, a data-separation
      theorem of depth k is resolved for up to >breadth< number
      of free variables in the current proof-state.

      NOTE: At present, the generation of data-separation
      lemmas is limited to "real" free variables and will
      not work for parameters of a test specification. Further,
      the proof-state is assumed to have only one goal.
    *)



   val  gen_test_data     : theory -> clasimpset -> string  -> testenv -> testenv
   (* "gen_test_data name env" constructs ground type instances of
      test theorems associated to "name". In a further step, 
      "abstract" or local test data were resolved into the resulting test 
      theorems (if available in the testenv under "name"), remaining
      meta variables were instantiated by a "random_solver"
      that attempts to generate an instance and tries to solve
      the constraints of a goal up to "iterations" times.
  
      The resulting simplified and "solved" test data theorems
      were stored again in the test environment.
    *)
      
   val get_test_data      : testenv * string  -> thm list
   (* get_test_data (te,name) extracts the test judgements
      (i.e. the premisses of test data statement, that are not
      test hypothesis.). *)

   val get_test_hyps      : testenv * string -> thm list
   (* get_test_data (te,name) extracts the test hypothesis
      of a !!test data statement!!, not a test theorem. *)



   val animate            : theory -> string * testenv -> unit 
   (* ground test data statements may be animated, if "P(progname x1 .. xn)"
      is executable. "animate" presents test data statements by 
      unfolding "progname x1 .. xn" and simplifying on a stepwise
      basis. (* for the moment, it's just asm_simp_tac with the
      standard evaluation trace, in the future its more or less
      the same except with slighly improved output *)
    *)

   val gen_testscript     : theory -> string -> string -> string -> testenv -> theory 
   
   val export_test_data   : theory ->string -> string -> testenv -> theory 

   val get_depth          : testenv -> int 
   val get_breadth        : testenv -> int
   val get_gen_prelude    : testenv -> bool 
   val get_gen_wrapper    : testenv -> bool 
   val get_gen_toString   : testenv -> string 

   val depth_update       : int  -> testenv -> testenv
   val setup_code_update  : string  -> testenv -> testenv
   val dataconv_code_update  : string  -> testenv -> testenv
   val breadth_update     : int  -> testenv   -> testenv
   val bound_update       : int  -> testenv   -> testenv
   val case_breadth_update: int  -> testenv -> testenv
   val iterations_update  : int  -> testenv -> testenv
   val gen_prelude_update : bool  -> testenv -> testenv
   val gen_wrapper_update : bool  -> testenv -> testenv
   val gen_toString_update : string  -> testenv -> testenv
   val type_range_bound_update : int  -> testenv -> testenv
   val type_candiates_update   : typ list -> testenv -> testenv
						      
end
 

(* structure TestGen : TESTGEN = *)

structure TestGen (*  : TESTGEN *) =
struct

open HOLogic;



(* ************************************************************************* *)
(* access to datatype package ...                                            *)
(* ************************************************************************* *)

fun isDataType thy (Type(tn,_)) =
               is_some(Symtab.lookup(DatatypePackage.get_datatypes thy) tn)
   |isDataType thy _ = false;


fun get_exhaust_of type_name thy =
    #exhaustion(the(Symtab.lookup(DatatypePackage.get_datatypes thy) type_name));
fun get_nchotomy_of type_name thy =
    #nchotomy(the(Symtab.lookup(DatatypePackage.get_datatypes thy) type_name));


(* ************************************************************************* *)
(* term computations ...                                                     *)
(* ************************************************************************* *)

fun max_depth (Free h) = (0,[h])
   |max_depth (Const h)= (0,[])
   |max_depth (Abs(s,tt,t)) = max_depth t
   |max_depth (s $ t) = let val (k,S) = max_depth (s)
                            val (k',S') = max_depth (t)
                        in  if k > k'+1 then (k,S)
                            else if k = k'+1 then (k, S @ S')
                                 else (k'+1, S')
                        end
   |max_depth (_) = (0,[]);


(* relies on definitions in Testing.thy ... *)
fun is_thyp (Const ("Trueprop",_) $ (Const ("Testing.THYP",_) $ _)) = true
   |is_thyp _   = false;

(*
fun strip_implies (Const("==>",_) $ S $ T) = let val (H,R) = strip_implies T
                                             in  (S::H,R) end
   |strip_implies S = ([],S);
 *)

fun mk_listT T = Type ("List.list", [T]);


fun is_fun_Type (Type("fun",_)) = true
   |is_fun_Type _               = false



fun calc_constr_list tgt descr = 
let fun DtTFree_tvar x = (DatatypeAux.dest_DtTFree x, HOLogic.typeS)
    val sorts          = Library.flat (map (fn (_,(_,s,_)) => map DtTFree_tvar s) descr)
    val recTs          = DatatypeAux.get_rec_types descr sorts;
    val used           = foldr add_typ_tfree_names [] recTs;
    val newTs          = Library.take (length (descr), recTs);
    val (_,(_,insts,tgt_constrs)) = hd(filter (fn (_,(n,_,_)) => n = tgt) descr)
    val T              = hd(filter (fn (Type(n,_)) => n = tgt) newTs)
    val typ_of_dtyp    = DatatypeAux.typ_of_dtyp descr sorts
    val constr_decls   = map (fn (cname, cargs) =>
                                 Const(cname, map typ_of_dtyp cargs ---> T))
                             (tgt_constrs)
in  (map DatatypeAux.dest_DtTFree insts, constr_decls) end;

(* test:
   val descr = 
   calc_constr_list s descr;
 *)


fun random_term' w n max thy cod_term_tab (Type(s,S)) =
(* w => the weight on the actual level, initial value 1
   n => level counter, inital value 0 
   max => maximal allowed number of levels
*)
       (case Symtab.lookup cod_term_tab s of
         NONE   => (* default random term generator ... 
                      TODO : should do also something for functions ... *)
                   (case s of 
		      "Int.int" => mk_number intT (IntInf.fromInt(random_range 0 10))
                    | "set" => Const("List.set",dummyT) $
                                    (random_term' w n max thy cod_term_tab (Type("List.list",S)))
                    | _     => let 
		    	           val descr = #descr(the(Symtab.lookup
                                                  (DatatypePackage.get_datatypes thy) s))

		    	           val (insts,constrs) = calc_constr_list s descr
				   
				   val weighed_constrs = let
				       fun ar args = (length 
							      (filter (fn t => 
									  case t of 
									      DatatypeAux.DtRec _ => true 
									    | _ => false ) args) ) 
				       val constr_arity_list =  map (fn (f,args) => (f,(ar args))) 
								  ( Library.flat (map (#3 o snd) descr))
				   in
				       map (fn (f,a) => if a = 0 then (1,f) else (a * w,f)) 
					   constr_arity_list
				   end
			      
				   val  weighed_constrs = if (n >= max) 
							  then filter (fn (w,f) => w =1) weighed_constrs
							  else weighed_constrs
			           fun weight_of t = fst(hd ((filter (fn (w,ty) => ty=t)) weighed_constrs)) 

				   fun frequency' xs = let 
				       val max = foldl Int.max 0 (map fst xs);
				       val xs' = map (fn (x,a) => (max-x+1,a)) xs
				   in
				       frequency xs'
				   end
				   (* the default is a random bias towards constants *)
				   val constr = frequency' weighed_constrs
                                   val Const(h,ty) = hd (filter (fn Const(h,ty) => h = constr) constrs) 
				   val w          = weight_of h 
		    	           val ty_binds   = insts ~~ S 
		    	           fun ty_inst (s,_) = snd(the(Library.find_first 
				                           (fn (x, _) => x = s)
				                           (ty_binds)))
		    	           val instantiated_ty = map_type_tfree ty_inst ty
		    	           val const_head = Const(h,instantiated_ty)
		    	           val arg_ty = binder_types instantiated_ty
		    	       in  list_comb(const_head,
                                             (map(random_term' w (n+1) max thy cod_term_tab)(arg_ty)))
                               end 
                               handle _=>error("Cannot generate random value for type:" ^s))
       | SOME R => R S)
   |random_term' _ _ _ _ _ _ = error "Internal error in random_term: type not ground";


fun random_term thy cod_term_tab typ = random_term' 1 0 10000 thy cod_term_tab typ  

(* test section:

val ttt = [HOLogic.intT,HOLogic.unitT,HOLogic.boolT,
           HOLogic.mk_setT HOLogic.intT, 
           mk_listT HOLogic.intT]
map (random_term Symtab.empty) ttt;

 *)


(* complex type instantiations ... *)

(* groundT_thm : testenv -> thm -> thm list *)


fun groundT_thm thy bound candTs thm =
    let val thm   = Thm.varifyT thm
        val tvars = term_tvars (list_comb(concl_of thm,prems_of thm))
        fun ctyp_of_sort x so t = if Sign.of_sort thy (t,so) 
                                  then (SOME(ctyp_of(thy)(TVar(x,so)),ctyp_of thy t) 
				             handle _ => NONE)
                                  else NONE
        fun prodS [] B = []
           |prodS (a::R) B =  (map(fn x => a::x) B) @ (prodS R B)

        fun compute_instsT [] = [[]]
           |compute_instsT ((x,so)::R) = prodS (List.mapPartial (ctyp_of_sort x so) candTs)
                                               (compute_instsT R)
    in 
	map (fn x => instantiate (x,[]) thm)
	    (Library.take(bound,compute_instsT tvars))
    end;
    
(* test

 val thm' = groundT_thm thy 
                        (#type_range_bound(rep_testenv te)) 
                        (#type_candicates(rep_testenv te)) 
                        thm

 *) 


(* *********************************************************************** *)
(*                                                                         *)
(* SOME more universal tacticals -> tctical.ML                             *)
(*                                                                         *)
(* *********************************************************************** *)

fun TRY' tac n = TRY(tac n) (* standard combinator,
                               not implemented so far in Isabelle *)


fun COND' testfun thenf elsef x prf = 
      if testfun x prf then thenf x prf else elsef x prf;


fun ALLGOALS_before k tac st =
  let fun doall 0 = all_tac
        | doall n = tac(n) THEN doall(n-1)
      val m = nprems_of st
  in  if k <= 0 orelse k > m 
      then all_tac st 
      else doall k st  
  end;




(* *********************************************************************** *)
(*                                                                         *)
(* Goal-Package with goal stack for Enumerative Deductions of              *)
(* proof-states. Alternative Derivations can be "spawned off" and were     *)
(* managed in a stack of stacks.                                           *)
(*                                                                         *)
(* *********************************************************************** *)


val derivative_stack_stack = ref([]:(int*thm)list list);
(* or: (int*thm)list seq ??? *)

fun dss_init ()       = (derivative_stack_stack:=[]);
fun dss_push S        = (derivative_stack_stack:=
                             (S::(!derivative_stack_stack)));
fun dss_pop  ()       = (let val S::R = !derivative_stack_stack
                         in  derivative_stack_stack:=R; S end);
fun dss_enq  (k,thm)  = dss_push((k,thm)::dss_pop());
fun dss_mt   ()       = (null(!derivative_stack_stack));
fun dss_mtq  ()       = (case !derivative_stack_stack of
                           []    => true
                         | []::R => true
                         | _     => false);

(* val refine_thm2 : tactic -> thm -> thm * (int * thm) list

   core of the alternative goal package.

   converts a thm to a list of "tagged derivatives" by applying
   the tactic (which may have side-effect on stack-stack) 
   and reports the resulting top of stackstack.
   Note that free variables remain free and were not
   converted to meta-variables.
 *)
val hack3 = ref (allI)

fun refine_thm2  tac thm =
    let 
	val _ = hack3 := thm
	val sg     = theory_of_thm thm
        val ct     = cterm_of sg (concl_of thm)
        val k      = nprems_of thm
        val frees  = term_frees (list_comb(concl_of thm,prems_of thm))
        val frees2 = map(fn(m as Free(x,t)) => (cterm_of sg (Var((x,0),t)),
                                                cterm_of sg m)) frees;
        (* uresult turns free's to vars due to standardization.
           with frees2, this substitution can be undone. *)
    in  dss_push [];
      (*  IsarCmd.disable_pr(); *)
       OldGoals.goalw_cterm [] ct;
         by(rtac thm 1);
        (by(tac)
         handle _ => ()); (*catch all exceptions*)
        (instantiate ([],frees2) (uresult()), ((*enable_pr();*)dss_pop()))
    end
 

(* refine_thm is equivalent refine_thm2 but forgets the labels.*)
fun refine_thm tac thm = fst(refine_thm2 tac thm);


(* val INIT_SS : tactic
   inits stackstack, always succeeds *)
fun INIT_SS thm      = (dss_push []; all_tac thm)

        
(* val SPAWN_SS : int -> tactic
   enstacks current state into topmost stack,
   tags it with "label" n. *)
fun SPAWN_SS n thm   = (dss_enq (n,thm); all_tac thm);


(* val CATCH_SS : tactic 
   fails if stackstack mt, and succeeds otherwise.
   Returns pops toptop element and returns it 
   when succeeding. *)
fun CATCH_SS thm     = 
       if dss_mtq() 
       then no_tac thm
       else let val (_,a)::R = dss_pop()
            in  dss_push R; all_tac a end

(* val CATCHJOIN_SS : tactic 
   Similar CATCH_SS, but fuses (rest of) topstack into
   next stack of stackstack. 
   (PRE provided this exists. length(!derivative_stack_stack) > 1 )*)
fun CATCHJOIN_SS thm = 
       if dss_mt() 
       then no_tac thm
       else let val R = dss_pop()
                val S = dss_pop() 
            in  case R of
                  []       => (dss_push (R@S); no_tac thm)
                | (_,a)::R => (dss_push (R@S); all_tac a) 
            end



(*  
fun COLLECTS_SS : Tactical.tactic list -> int -> Tactical.tactic 
    PRE  not (dss_mt()) 
    applies tactic list to proof-state. Any success is stacked
    in top stack of stackstack. Always fails.
*)
fun COLLECTS_SS [] n      = no_tac
   |COLLECTS_SS (t::tl) n = (((t n) THEN (SPAWN_SS (n-1)) THEN no_tac) 
                             ORELSE (COLLECTS_SS tl n));


(* COLLECT_SS : (int -> tactic) * (int -> tactic) -> int -> tactic
   Inits stack and performs both tac1 and tac2 on implicit goal
   "n".
   possible outcomes:
   - failure, mt stack     if both tactics fail
   - success, mt stack     if one of the tactics succeed
   - success, nonmt stack  if both tactics succeed. *) 
fun (tac1 COLLECT_SS tac2) = 
    ((K INIT_SS) THEN'
     COLLECTS_SS[tac1,tac2]) ORELSE' (K CATCH_SS);


(* val CLOSURE : (int -> tactic) list -> int * thm -> (int * thm) list
   performs a list of goal-dependent tactics for all goals
   and APPLIES TO ALL NEW DERIVATIVES produced by this step
   *BEFORE*

 *)
  
val hack1 = ref (allI)
val hack2 = ref (allI)

fun CLOSURE [] (n,thm) = [(n,thm)] 
   |CLOSURE taclist (n,thm) =
    let 
	val _ = hack1 := thm
	val TRY_CATCHJOIN   = TRY' (K CATCHJOIN_SS)
                              (* catchjoin uses a derivative
                                 to continue, provided there exists
                                 one. Otherwise ("TRY") the original
                                 proof state is returned *)
        val (nthm, derivs)  = refine_thm2 
                                (ALLGOALS_before n
                                   ((K INIT_SS) THEN'
                                    ((COLLECTS_SS taclist)
                                     ORELSE' TRY_CATCHJOIN)))
                                (thm)
                              (* produces a new stack on top of stack-stack,
                                 runs COLLECT_SS (which always fails)
                                 pops the derives and joins them into 
                                 previous stack 
                                 *)
	val _ = hack2 := nthm
    in  (0,nthm) :: (Library.flat(map (CLOSURE taclist) derivs)) end;


fun CLOSURE' taclist thm = map snd (CLOSURE taclist (nprems_of thm,thm));



(* *********************************************************************** *)
(*                                                                         *)
(* special tacticals for TestGen                                           *)
(*                                                                         *)
(* *********************************************************************** *)


(* PRE n>0 and n <= nprems_of thm *)
fun IS_THYP n thm = is_thyp(List.nth(prems_of thm, n-1)) 

fun ALLNONTHYPGOALS_before tac n = 
    ALLGOALS_before n (COND' IS_THYP (K all_tac) tac)

fun ALLNONTHYPGOALS tac thm =
    (ALLNONTHYPGOALS_before tac (nprems_of thm) thm)


(* *********************************************************************** *)
(*                                                                         *)
(* basic tactics ...                                                       *)
(*                                                                         *)
(* *********************************************************************** *)

fun mp_fy n = (REPEAT(etac rev_mp n));

fun all_ify excpt n state = 
    let val subgoal    = List.nth(prems_of state,n-1)
        val params     = Logic.strip_params subgoal
        val frees      = map dest_Free (term_frees subgoal)
        fun sbs(n,t)   = res_inst_tac [("x",n)] spec
        val vars       = filter (fn x => not((fst x) mem excpt))
                                (params@frees)
    in  (if null vars then no_tac 
         else foldr(fn(x,ta) => (sbs x n) THEN ta)
                   prune_params_tac vars) 
        (state) 
    end


fun convert_goals_to_metahyps (test:(term -> bool)) thm = 
    let val sg = theory_of_thm thm
    in  map (fn prop => Thm.forall_elim_vars 0 (assume (cterm_of sg prop)))
            (filter test (prems_of thm))
    end;


fun instantiate_tac insts n thm =
    let val goal = cterm_of (theory_of_thm thm) (List.nth(prems_of thm,n-1))
    in  (rtac(instantiate ([],insts)(trivial goal)) n) thm end



val uni_True = OldGoals.prove_goalw (theory "Main") [] "x = x ==> True"
  (fn prems => [cut_facts_tac prems 1,auto_tac (clasimpset())]);


fun var_hyp_subst_tac no thm =
    let val prms = Logic.strip_imp_prems(
                           List.nth(prems_of thm,no-1));
        fun is_eq (Const ("op =", _) $ (Var _) $ _) = true
           |is_eq (Const ("op =", _) $ _ $ (Var _)) = true
           |is_eq _ = false
        val i = Library.find_index(fn (Const("Trueprop",_)$X) => is_eq X) 
                                  (prms)
        fun mks thm = if i >= 0 
                      then ((rotate_tac i no) THEN dtac uni_True no) thm
                      else no_tac thm
    in ((REPEAT mks) 
        THEN 
        (full_simp_tac (simpset_of(theory"Main")) no)) thm 
    end;



(* *********************************************************************** *)
(*                                                                         *)
(*  special tactics for TestGen                                            *)
(*                                                                         *)
(* *********************************************************************** *)


fun mk_Uniformity_THYP term =
    let fun convAll(Const("All",ty)$Abs(n,ty',tt)) = 
                      (Const("Ex",ty)$Abs(n,ty',convAll tt))
           |convAll tt = tt;
        val tt   = dest_Trueprop term
        val thyp = Const ("Testing.THYP", [boolT]--->boolT) $
                        (Const ("op -->", [boolT,boolT]--->boolT)
                         $ (convAll tt) $ tt)
    in  mk_Trueprop thyp end


fun unifomityI_tac thy prog n state =
    let fun prem_of state = List.nth(prems_of state, n-1)
        fun core_tac state = subgoal_terminst_tac [] 
                                (mk_Uniformity_THYP (prem_of state)) 
                                (n) (state)
	val count = ref 0;
	fun nwstr() = (count:=(!count)+1; "?X"^(Int.toString(!count))^"X")
    in (if "Testing.THYP" mem (term_consts(prem_of state))
        then no_tac
        else  EVERY[mp_fy n, all_ify [prog] n, 
                    core_tac, etac (PureThy.get_thm thy "THYP_app1") n, 
                    REPEAT(fn t => (res_inst_tac [("x",nwstr())] exI n)(t)), 
		    strip_tac n]) 
       state
    end


(* lemma *)
val swap_accumulate = prove_goal (theory "Main") "[| P ; ~P | Q |] ==> Q"
    (fn prems => [cut_facts_tac prems 1, auto_tac (claset(),simpset())])


(* ************************************************************************* *)
(* data separation lemma generation ...                                      *)
(* ************************************************************************* *)

fun prems_hyp_subst_tac thy name H1 =  
let fun stripC (Const("==>",t) $ H1 $ (Const("==>",_)$ _ $ H2))=
                   Const("==>",t) $ H1 $ H2
       |stripC (Const("all",t) $ (Abs(s,t',H))) = 
                   Const("all",t) $ (Abs(s,t', stripC H)); 
    fun stripD name (Const("all",t) $ (Abs(s,t',H))) = 
                   if s = name then incr_bv (~1, 0,stripC H)
                   else Const("all",t) $ (Abs(s,t', stripD name H)); 
    val H2     = stripD name H1;
    val cgoal1 = cterm_of ( thy) (Logic.mk_implies(H1,H2));
    val cgoal2 = cterm_of ( thy) (Logic.mk_implies(H2,H1));
    val pp1    = OldGoals.prove_goalw_cterm [] cgoal1
                 (fn [prems] => [rtac prems 1, atac 1, rtac refl 1]);
    val pp2    = OldGoals.prove_goalw_cterm [] cgoal2
                 (fn [prems] => [rtac prems 1, atac 1]);
in  equal_intr pp1 pp2 end;


val changed = ref false;


fun refine_data_separation_lemma thy clasimp bal prem_no cands thm =
(* - bal controls of the refinement is in balancing mode
     (only parameters whose occurrence in the term is
      has strictly less than maximum height)
     or not. 
   - prem_no indicates the number of the premise
   - cands is the global list of candidates.
     In balancing mode, parameters in maximal depth
     position were eliminated.
 *)
let (* val thy    = theory_of_thm thm *)
    val prem   = List.nth(prems_of thm, prem_no - 1)
    val params = Logic.strip_params(prem)

    val cands' = if bal then 
                    let val term   = HOLogic.dest_Trueprop(hd
                                            (Logic.strip_assums_hyp prem))
                        val (k,mp) = max_depth(Term.subst_bounds
                                            ((rev(map Free params)),term))
                    in filter(fn p => forall(fn p'=> not(p=p')) (mp)) cands end
                 else cands;

    val var_no = find_index (fn(s,t)=> exists(fn (name,ty)=> 
                                             s = name andalso t = ty) cands')
                             params
    val (name,ty)  = if var_no > ~1 then List.nth(params,var_no)
                     else raise (ERROR("no param found"))
    val Type(tn,_) = ty;
    val m          = get_exhaust_of tn thy;
    val n          = m RSN (prem_no,thm);

    (* generates a projection function % x1 ... xn. xk of the form
       "%x x y . x" in order to instantiate the freshly generated
       meta-variable y1 in the RSN step above. y1 will occur
       lifted over the parameters param and can therefore be grounded
       by this projection *) 
    val proj  = concat(["%"] @ (replicate  (var_no+1) "x ") @ 
                       (replicate  (length(params) - var_no - 1) "y ") @ 
                       [". x"])
    val p     = read_instantiate [("y1",proj)] n; (* always y1 ? *)

    (* Now, why is the rest so complicated ?
       val p' = rule_by_tactic (ALLGOALS hyp_subst_tac) p; 
       destroys everything, on the other hand, 
       rule_by_tactic (prune_params_tac) p'; has no effect at all. 
       Therefore, the following rewrite  plus cleanup is necessary *)
    val p'    = rule_by_tactic (ALLGOALS (asm_full_simp_tac
                                          (snd clasimp))) p;
 (*   val simps = Library.mapfilter  *)
 (*               (try(prems_hyp_subst_tac (theory_of_thm p') name))  *)
 (*               (prems_of (fst(freeze_thaw(p')))); *)
    val simps = List.mapPartial 
                (try(prems_hyp_subst_tac thy name)) 
                (prems_of (fst(freeze_thaw(p'))));
  
    val ()    = (changed:=true)
in  rule_by_tactic (ALLGOALS (asm_full_simp_tac 
                                  (HOL_ss addsimps simps))) p'
end;



(* This is the core procedure of the data_separation_lemma generation, and,
   in a way, the core of the test case generator itself.

   It proceeds as follows: in a given theorem thm (usually an exhaustion lemma 
   itself), the premisses (having the form !!x1..xn. A x1 .. xn ==> B in 
   general) were selected and "refined" in a "sweep". This means that in a 
   sweep, for each premisse an xi is selected (which must have a data-type, 
   say ti) and the exhaustion theorem of ti is forward resolved into this 
   premisse; cleaning it up (by prems_hyp_subst_tac) also erases xi as 
   parameter of this premisse. 
   In a next sweep, another xj may be selected and so forth, up to the moment 
   were no parameters remained in the premisse that existed before the initial 
   sweep (this the list parameters is the "cand" (candidates) x1 .. xj for 
   all premisses which precomputed.)

   Then, the candidate list may be recomputed and the process repeated up to 
   n times. Since any group of sweeps erases one generation of parameters, 
   the parameter n corresponds to "depth+1" of the data term occuring in the 
   premisse increases, i.e. from:

   [| ?y = Leaf ==> ?P; 
      !!a tree1 tree2. ?y = Node a tree1 tree2 ==> ?P |] 
   ==> ?P

   we get:

   [| ?y = Leaf ==> ?P; !!a. ?y = Node a Leaf Leaf ==> ?P;
      !!a aa tree1 tree2a. ?y = Node a Leaf (Node aa tree1 tree2a) ==> ?P;
      !!a aa tree1a tree2a. ?y = Node a (Node aa tree1a tree2a) Leaf ==> ?P;
      !!a aa tree1a tree2a ab tree1 tree2b.
            ?y = Node a (Node aa tree1a tree2a) (Node ab tree1 tree2b) ==> ?P 
   |] ==> ?P

   This picture is slightly complicated by the fact that Isabelle does not
   change the parameters, but may reintroduce parameter erased by a previous 
   sweep during the forward resolution step. Thus, the procedure may loop.

   Instead of not erasing the parameters during lemma-refinement (this
   leads to very large liftings of the meta-variables over these
   parameter lists and is thus expensive), we chose to use 
   a particular test for "maximal" parameters in order to avoid this problem.

 *)  
fun nrefine_data_separation_lemma thy clasimp 0 thm = thm
  | nrefine_data_separation_lemma thy clasimp n thm =
    let 
      val thy           = theory_of_thm thm; (* HACK ISAR problem *)
      val cands         = distinct (op=) (filter (fn(_,ty)=> isDataType thy ty) 
                                           (Library.flat(map Logic.strip_params 
							     (prems_of thm))))
      fun rds b(prno,t) = refine_data_separation_lemma thy clasimp 
                                                       b prno cands t 
                            handle _ => t
      val ()            = (changed:=false)
      fun sweep b th    = foldr (rds b) th (1 upto (length(prems_of th)))
      val thm1          = sweep false thm       (* debalancing phase *)
      fun repeat_till_stable thm = let val ()   = (changed:=false)
                                     val thm' = sweep true thm
                                   in  if !changed 
                                       then repeat_till_stable thm' 
                                       else thm'
                                   end
    in  nrefine_data_separation_lemma thy clasimp
          (n-1) 
           (if  !changed 
            then thm1                     (* debalancing phase had no effect *)
            else repeat_till_stable thm1) (* balancing phase *)
    end;

(* NOTE: Currently, this core routine is restricted to
   "real free" variables; parameters will not be
   considered as a simple means to avoid looping ... *)
fun data_separation_tac thy clasimp depth  no state = 
    let 
    (*  val thy    = theory_of_thm state *)
        val prem   = List.nth(prems_of state, no - 1)
        val params = Logic.strip_params(prem)
        val term   = HOLogic.dest_Trueprop(Logic.strip_assums_concl prem)
        val frees  = map dest_Free (term_frees term)
(*      val cands  = filter (fn (_,ty) => isDataType thy ty) (params@frees) *)
        val cands  = filter (fn (_,ty) => isDataType thy ty) (frees)
        fun ds_thm tn = nrefine_data_separation_lemma thy
                          (clasimp)(depth)(get_exhaust_of tn thy)
    in (case cands of 
          [] => no_tac
        | (na,Type(tn,_))::_ => let val data_sep_thm = ds_thm tn
                                    val brchs = length(prems_of data_sep_thm)
                                in  (res_inst_tac [("y",na)] data_sep_thm no)
                                    THEN
                                    (res_inst_tac [("x",na),
                                                   ("k",Int.toString depth)] 
                                                  (PureThy.get_thm thy 
								   "THYP_E_reg")
                                                  (no + brchs - 1))
                                    THEN
                                    (thin_tac (na^"= ?X") (no+brchs))
                                end)
       (state)
    end;




(* ************************************************************************* *)
(* normal form computation ...                                               *)
(* ************************************************************************* *)

fun to_TNF thy prog term no state =
    let
	val thy     = theory_of_thm state;
	val prems   = Logic.prems_of_goal term no
	val concl   = Logic.concl_of_goal term no

        val prems'  = filter (fn t => prog mem (add_term_names(t,[]))) prems;
        val concl'  = filter (fn t => prog mem (add_term_names(t,[]))) [concl];
	fun subst t = (Thm.cterm_of thy (Var(("P",0),HOLogic.boolT)),
                       Thm.cterm_of thy t);
        (*  swap_tac : tactic * term -> tactic; takes one term containing 
            "prog" and swap it into the conclusion as disjoint *)
        fun swap_tac (tm,tc) = tc THEN 
                               (eres_terminst_tac 
                                  []
                                  [subst (HOLogic.dest_Trueprop tm)]
                                  swap_accumulate 
                                  no)
        (* rev_tac swaps term "a" into conclusion *)
        fun rev_tac thy a = eres_terminst_tac 
	                          [] 
                                  [subst (HOLogic.dest_Trueprop a)] 
                                  (PureThy.get_thm thy "rev_notE") 
                                  no
    in ((case (prems',concl') of
          ([],[])   => error ("Subgoal " ^Int.toString(no)
                             ^" can not be converted into TNF.")
          (* case that conclusion does not contain "prog", but several
             assumptions do: pick one of them, revert first *)
         |(a::R,[]) =>(rev_tac thy a
                       THEN
                       (foldl swap_tac all_tac R))
         |(R,_)     => (foldl swap_tac all_tac R)
        )
        THEN
         (simp_tac HOL_ss no)) state
    end;


(* TODO: adb *)
(*
fun normalize_TNF prog top =
    let val prems = prems_of top
        val mrs   = prems ~~ (1 upto (length prems))  
    in  foldl (fn(arg,t)=> t THEN to_TNF prog arg) all_tac mrs (top)
    end
 *)
fun normalize_TNF thy prog top =
    let val mrs   = (length (prems_of top)) downto 1
    in  foldl (fn (arg,t) => t THEN to_TNF thy prog (prop_of top) arg) 
              all_tac 
	      mrs 
	      top
    end


(* TODO: adb *)
val minimize_TNF  = distinct_subgoals_tac
(* This should do:
   - sorting assumptions in all subgoals
   - sorting parameters of subgoals
   - sortieren der clauses - 
     weniger asms zuerst (defer_tac)
   - eliminieren von assumptions subsumed in other clauses
     [| P;R |] ==> A;
     [| P;Q;R  |] ==> A
     ------------------
     [| P;R |] ==> A;
     [| P;R  |] ==> A
     durch thin_tac
   - eliminieren von subsumed or's,
     a particular weakness of our procedure so far.
     [| P;R |] ==> A;
     [| ~P;Q |] ==> B
     [| R; Q|] ==> A|B;
     ------------------
     [| P;R |] ==> A;
     [| ~P;Q |] ==> B
    - finally distinct_subgoals_tac
*)


(* ************************************************************************* *)
(* test case generation ...                                                  *)
(* ************************************************************************* *)


(* orig version *)
(*
fun gen_test_case_tac thy clasimp depth breadth prog  =
    EVERY[ALLGOALS (asm_full_simp_tac (snd clasimp)),
          safe_tac (claset() addss simpset()),
          normalize_TNF thy prog,
          minimize_TNF,
          ALLGOALS (fn n => TRY(unifomityI_tac prog n))];
*)
val trace_test = ref false; 

fun gen_test_case_tac thy clasimp depth breadth prog  =
    let fun standard thy k b = 
            if !trace_test
            then  EVERY[print_tac ("\n"^b^"Enter\n"),
                        ALLGOALS(asm_full_simp_tac (snd clasimp)),
                        print_tac ("\n"^b^"Simp\n"),
                       (* if k > 0 
                        then TRY(safe_tac(fst clasimp))
                             (* HCNF computation may fail, but
                                since a TNF has been computed,
                                we report overall success *)
                        else safe_tac(fst clasimp)
                             (* core of gen_test never run 
                                successfully; failure reported *),
                        *)
                        TRY(safe_tac(fst clasimp)),
                        print_tac ("\n"^b^"HCN\n"),
                        normalize_TNF thy prog,
                        print_tac ("\n"^b^"TNF\n"),
                        minimize_TNF,
                        print_tac ("\n"^b^"MTNF\n")]
            else  EVERY[ALLGOALS(asm_full_simp_tac (snd clasimp)),
(*                        if k > 0 
                        then TRY(safe_tac(fst clasimp))
                             (* HCNF computation may fail, but
                                since a TNF has been computed,
                                we report overall success *)
                        else safe_tac(fst clasimp)
                             (* core of gen_test never run 
                                successfully; failure reported *),
  *)                      TRY(safe_tac(fst clasimp)),
                        normalize_TNF thy prog,
                        minimize_TNF]
        val dst = data_separation_tac thy clasimp depth  


	(*True if the two theorems have the same prop field, ignoring hyps, der, etc.*)
	val eq_thm_prop = op aconv o pairself Thm.full_prop_of;

        fun main thy k st = 
			(if k < breadth 
                     (* not yet all frees per goal less breadth exhausted*)
                        then 
			(ALLGOALS(fn n => TRY(dst n)))
                          (* try exhaustion per goal *)
                             THEN COND(fn st' => not (eq_thm_prop(st, st')))
                                  (* something 's changed *)
                                     ((ALLGOALS (fn n => TRY(hyp_subst_tac n))) 
                                     THEN standard thy k "1" 
                                     THEN (main  thy (k+1))) 
                                   (standard thy k "2")
		     else 
		     	standard thy  k "0" (* run out of breadth *)) (st)
    in  (ALLGOALS((asm_full_simp_tac (snd clasimp))))
         THEN (if !trace_test then print_tac ("\n pre-simplification \n") 
	      else all_tac)
        THEN (main  thy 0)
	THEN (if !trace_test then print_tac ("\n main completed \n") 
	      else all_tac)
        THEN (ALLGOALS(fn n => TRY(unifomityI_tac thy prog n)))
	THEN (if !trace_test then print_tac ("\n main uniformity NF \n") 
	      else all_tac)
    end;


(* ************************************************************************* *)
(*                                                                           *)
(* testdata generation        ...                                            *)
(*                                                                           *)
(* ************************************************************************* *)

  (* *********************************************************************** *)
  (* Hierarchical Solver                                                     *)
  (* *********************************************************************** *)


val deriso = prove_goal (theory "Main") "[| P; P |] ==> True"
    (fn prems => [cut_facts_tac prems 1,  
                  auto_tac (claset_of (theory "Main"),
                            simpset_of (theory "Main"))]);


fun test_case_solve_tac ss thm = 
    let fun list2string f S = "["^(foldr(fn(x,y)=>(f x)^","^y) "" S)^"]"
        fun str n = Int.toString n
        fun dump n S = writeln((str n) ^ " " ^
                               concat(map(fn T=>list2string(str o fst)T)S))
    in  COND' IS_THYP 
              (K  no_tac) (* crucial for its use in CLOSURE *)
              (fn n =>(dump n (!derivative_stack_stack); 
                       (dtac (thm RS deriso) n) THEN (full_simp_tac ss n)))
    end


  (* *********************************************************************** *)
  (* Random Solver                                                           *)
  (* *********************************************************************** *)

fun random_inst_tac thy cod_tab ss iterations n thm' =
    let val sg    =   thy
        val prems = prems_of thm'
        val m     = length prems
	val _     = tracing ("Random solving subgoal "^Int.toString(n)^
                                                " of "^Int.toString(m))  
        val prem  = List.nth(prems,n-1)
	val k     = length (Logic.strip_imp_prems(prem)) 
        val goal  = trivial(cterm_of(sg)(prem))
        val tvars = term_vars prem
        fun insts ()= map(fn(x as Var(s,t))=>
                            (cterm_of sg x,
                             cterm_of sg 
                                (random_term thy cod_tab t)))
                       (tvars)
	val cleanup  = (foldr (fn (_,tac) => (atac (n+1)) THEN tac)
			                     all_tac (1 upto k))
        fun ran_tac thm = (rtac (instantiate ([],insts())goal) n) thm

        val intro_RSF = (REPEAT(thin_tac "?X" n)) 
			    THEN (rtac (PureThy.get_thm thy "RSF_E") n)
			
      
	fun is_implies (Const ("==>", _) $ _ $ _) = true
	  | is_implies _ = false;

        fun solve itr= ran_tac
                       THEN 
                       (full_simp_tac ss n)  (* may erase goal n. --> failure *) 
                       THEN 
	               (fn thm'' => if (nprems_of thm'' < m+k) 
		                       (* goal erased, test case inconsistent *)
				       orelse
				       is_implies 
				                 (List.nth(prems_of thm'',n-1))
                                       (* constraint unsolved *)
	                            then if itr > 0 
				         then solve (itr-1) thm' 
					      (* do it again,Sam*)
				         else (* one final attempt without 
					         solving, and erases constraits by
                                                 RSF (random solve failure)
                                                 handled by the code generator. *)
				              (ran_tac 
                                               THEN cleanup
                                               THEN intro_RSF) (thm')
			            else (* constraints solved, now cleanup *)
				         cleanup thm'')
    in solve iterations thm'
    end;

 
fun random_solve thy cod_tab simp_ss iterations =
    refine_thm (ALLNONTHYPGOALS (random_inst_tac thy cod_tab 
                                                 simp_ss 
                                                 iterations));


  (* *********************************************************************** *)
  (* Normalizer                                                              *)
  (* *********************************************************************** *)
  (* motivation: type instantiation may make
     constraints of test-cases unsolvable. 
     Unsolvable cases should be removed before random-solving. 
     Since constraints with Mvars were not considered
     by simptac and arith_tac, they were substituted 
     against arbitrary free vars aforehand. *)

fun norm_tac thy ss n thm =
   let  val sg    =   thy
        val prems = prems_of thm
        val m     = length prems
        val prem  = List.nth(prems,n-1)
	val k     = length (Logic.strip_imp_prems(prem)) 
        val goal  = trivial(cterm_of(sg)(prem))
        val tvars = term_vars prem
        val insts = map(fn(x as Var((s,a),t))=>
                            (cterm_of sg x,
                             cterm_of sg (Free("XXX"^s^(Int.toString a),t))))
                       (tvars)
	val cleanup    = foldr (fn (_,tac) => (atac (n)) THEN tac)
			       all_tac (1 upto k)
        val ground_tac = (rtac (instantiate ([],insts) goal) n) 
   in  (
        ground_tac
        THEN 
        (full_simp_tac ss n)  (* may erase goal n. 
		                 --> test case unsolvable. *) 
        THEN 
	(fn thm' => if (nprems_of thm' < m+k) (* goal erased *)
	            then cleanup thm'
		    else no_tac thm )(* report failure *)
       )(thm)
   end
    

  (* *********************************************************************** *)
  (* core routine: gluing all local solvers together ...                     *)
  (* *********************************************************************** *)
  (* realizes the solver chain: first grounding,
     then erasing test cases that become unsolvable
     during grounding,
     then abstract test data resolution,
     then random solving. (could be enumeration instead).

     FUTURE: an arith-solver between abstract test data resolution and
     random solving. In particular sorting related tests would
     greatly profit from such a component.
   *)


fun gen_test_data_core thy name cdtb ss iters bd candTs thm tdstmts =
    let (* val ss           = simpset_of thy *) 
        val random_phase = random_solve thy cdtb ss iters
        val arith_phase  = fn thm => [thm]   (* not yet implemented *)
	val absdata_phase= CLOSURE'(map (test_case_solve_tac ss) tdstmts)
	val norm_phase   = refine_thm (ALLNONTHYPGOALS(TRY'(norm_tac thy ss)))
	val type_grounds = groundT_thm thy bd candTs thm 
    in  map (random_phase)
            (Library.flat(map (arith_phase)
                      (Library.flat(map (absdata_phase o norm_phase)
                                (type_grounds)))))
    end


(* ************************************************************************* *)
(*                                                                           *)
(* testenvironment management ...                                            *)
(*                                                                           *)
(* ************************************************************************* *)


type testinfo ={depth           : int, (* data structure depth *)
                breadth         : int, (* no of free's in testspec *)
                test_thm_tab    : thm Symtab.table, (* name to testthm *)

                bound           : int, (* global bound for data statements *)
                case_breadth    : int, (* no of test data per case,
                                          weakening uniformity *)
                iterations      : int, (* number of random attempts to solve a case *)

                type_range_bound: int,      (* effectively used elements 
                                               type grounding list *)
                type_candicates : typ list, (* type grounding list *)
 
		prelude         : bool,
		wrapper         : bool,
		toString        : string,
		setupCode       : string,
		dataconvCode       : string,
                cod_term_tab    : (typ list -> term) Symtab.table,
                                    (* random generators for terms,
                                       tab assigns to each type constuctor 
                                       name a generator that gets the
                                       instances of type constructor as args.*)
                abstest_data_tab: (thm list) Symtab.table, 
                                    (* assigns name to list of 
                                       data statements *)
                test_data_tab   : (thm list) Symtab.table};

datatype testenv   = Testenv of testinfo

val mt_testenv     = Testenv{depth           = 3,
                             breadth         = 1,
                             test_thm_tab    = Symtab.empty,

                             bound           = 200,
                             case_breadth    = 1,
                             iterations      = 25,

                             type_range_bound= 1,
                             type_candicates = [HOLogic.intT,HOLogic.unitT,
                                                HOLogic.boolT, 
                                                HOLogic.mk_setT HOLogic.intT, 
                                                mk_listT HOLogic.intT],
			     prelude         = true,
			     wrapper         = true,
                	     toString        = "",
                	     setupCode       = "",
                	     dataconvCode       = "",
                             cod_term_tab    = Symtab.empty,
                             abstest_data_tab= Symtab.empty,
                             test_data_tab   = Symtab.empty
};

fun rep_testenv (Testenv X) = X;
fun mk_testenv  (X:testinfo)= Testenv X

fun (Testenv {depth,breadth,test_thm_tab,
              bound,case_breadth,iterations,type_range_bound,
              type_candicates,prelude,wrapper,toString,setupCode,dataconvCode,
	      cod_term_tab, abstest_data_tab,test_data_tab}) 
	add_test_case 
    (name,thm)   =
    Testenv({depth=depth,
             breadth=breadth,
             test_thm_tab=Symtab.update(name,thm)test_thm_tab,
             bound=bound,case_breadth=case_breadth,
             iterations=iterations,type_range_bound=type_range_bound,
             type_candicates=type_candicates,prelude=prelude,wrapper=wrapper,toString=toString,setupCode=setupCode,
             dataconvCode=dataconvCode,cod_term_tab=cod_term_tab, 
             abstest_data_tab=abstest_data_tab,
             test_data_tab=test_data_tab});


(* inefficient !!! *)

fun(Testenv {depth,breadth,test_thm_tab,
             bound,case_breadth,iterations,type_range_bound,
             type_candicates,prelude,wrapper,toString,setupCode,dataconvCode,cod_term_tab, abstest_data_tab,
             test_data_tab}) 
    del_test_case 
    name  =
    Testenv({depth=depth,
             breadth=breadth,
             test_thm_tab=Symtab.make(filter (fn(x,_) => not(x=name)) 
                                             (Symtab.dest test_thm_tab)),
             bound=bound,case_breadth=case_breadth,
             iterations=iterations,type_range_bound=type_range_bound,
             type_candicates=type_candicates,prelude=prelude,wrapper=wrapper,
             toString=toString,dataconvCode=dataconvCode,setupCode=setupCode,
             cod_term_tab=cod_term_tab, abstest_data_tab=abstest_data_tab,
             test_data_tab=test_data_tab});




fun(Testenv {depth,breadth,test_thm_tab,
             bound,case_breadth,iterations,type_range_bound,
             type_candicates,prelude,wrapper,toString,setupCode,dataconvCode,
	     cod_term_tab, abstest_data_tab,test_data_tab}) 
    add_test_data 
    (name,thm) =
    Testenv({depth=depth,
             breadth=breadth,
             test_thm_tab=test_thm_tab,
             bound=bound,case_breadth=case_breadth,
             iterations=iterations,type_range_bound=type_range_bound,
             type_candicates=type_candicates,prelude=prelude,wrapper=wrapper,
	     toString=toString,setupCode=setupCode,dataconvCode=dataconvCode,
             cod_term_tab=cod_term_tab, 
             abstest_data_tab=Symtab.update
                           (name, thm::(case Symtab.lookup abstest_data_tab name of
                                         NONE => []
                                       | SOME S => S))
                            abstest_data_tab,
             test_data_tab=test_data_tab});


fun(Testenv {depth,breadth,test_thm_tab,
             bound,case_breadth,iterations,type_range_bound,
             type_candicates,prelude,wrapper,toString,setupCode,dataconvCode,
	     cod_term_tab,abstest_data_tab,test_data_tab}) 
       del_test_data 
    name  =
    Testenv({depth=depth,
             breadth=breadth,
             test_thm_tab=test_thm_tab,
             bound=bound,case_breadth=case_breadth,
             iterations=iterations,type_range_bound=type_range_bound,
             type_candicates=type_candicates,prelude=prelude,wrapper=wrapper,
	     toString=toString,dataconvCode=dataconvCode,setupCode=setupCode,
             cod_term_tab=cod_term_tab, 
             abstest_data_tab=Symtab.make(filter (fn(x,_) => not(x=name)) 
                                        (Symtab.dest abstest_data_tab)),
             test_data_tab=test_data_tab});


fun merge_testenv
   (Testenv{depth=dd,     breadth=db,
            test_thm_tab=tttab,   bound=b,
            case_breadth=cb,      iterations=i,
            type_range_bound=trb, type_candicates=tc,prelude=p,wrapper=w,toString=s,
            cod_term_tab=ctt,     abstest_data_tab=tdt,dataconvCode=dc,setupCode=sc,
            test_data_tab=tjt},
    Testenv{depth=dd',    breadth=db',
            test_thm_tab=tttab',  bound=b',
            case_breadth=cb',     iterations=i',dataconvCode=dc',setupCode=sc',
            type_range_bound=trb',type_candicates=tc',prelude=p',wrapper=w',toString=s',
            cod_term_tab=ctt',    abstest_data_tab=tdt',
            test_data_tab=tjt'}) =

    Testenv{depth         = Int.min(dd,dd'),
            breadth       = Int.min(dd,db'),
	    test_thm_tab  = Symtab.merge (Thm.eq_thm) (tttab,tttab'), 
            bound                 = Int.min(b,b'),
            case_breadth          = Int.min(cb,cb'),
            iterations            = Int.min(i,i'),
            type_range_bound      = Int.min(trb,trb'),
	    type_candicates       = distinct (op=) (tc@tc'),
	    prelude               = p andalso p',
	    wrapper               = w andalso w',
	    toString               = "" ,
	    setupCode             =sc^"\n"^sc',
	    dataconvCode          =dc^"\n"^dc',
            cod_term_tab          = Symtab.empty, 
                                    (* don't know what a senseful
                                       definition of override could
                                       be; therefore constraint to the
                                       simplest and most conservative
                                       one *)
            abstest_data_tab         = Symtab.merge Thm.eq_thms (tdt,tdt'),
            test_data_tab         = Symtab.merge Thm.eq_thms (tjt,tjt')};
   


fun thm_tab  te = #test_thm_tab (rep_testenv te)
fun absdata_tab te = #abstest_data_tab(rep_testenv te)
fun jdmt_tab te = #test_data_tab(rep_testenv te)
(* In the sequel we also use the term "test judgement" synonymously
   for "test data statement". *)

fun thm_tab_update
       (Testenv{depth=dd, breadth=db,
            test_thm_tab=tttab,   bound=b,
            case_breadth=cb,      iterations=i,
            type_range_bound=trb, type_candicates=tc, prelude=p,wrapper=w,toString=s,setupCode=sc,
            cod_term_tab=ctt,     abstest_data_tab=tdt,
		dataconvCode=dc,
            test_data_tab=tjt})
       (thm_tab) =
       (Testenv{depth=dd,     breadth=db,
                test_thm_tab=thm_tab, bound=b,
                case_breadth=cb,      iterations=i,
                type_range_bound=trb, type_candicates=tc,prelude=p,wrapper=w,toString=s,setupCode=sc,
                cod_term_tab=ctt,     abstest_data_tab=tdt,
		dataconvCode=dc,
                test_data_tab=tjt});

fun jdmt_tab_update
       (Testenv{depth=dd, breadth=db,
            test_thm_tab=tttab,   bound=b,
            case_breadth=cb,      iterations=i,
            type_range_bound=trb, type_candicates=tc,prelude=p,wrapper=w,toString=s,setupCode=sc,
            cod_term_tab=ctt,     abstest_data_tab=tdt,
		dataconvCode=dc,
            test_data_tab=tjt})
       (jdmt_tab) =
       (Testenv{depth=dd,     breadth=db,
                test_thm_tab=tttab,   bound=b,
                case_breadth=cb,      iterations=i,
                type_range_bound=trb, type_candicates=tc,prelude=p,wrapper=w,toString=s,setupCode=sc,
                cod_term_tab=ctt,     abstest_data_tab=tdt,
		dataconvCode=dc,
                test_data_tab=jdmt_tab});


fun depth_update (depth)
       (Testenv{depth=dd, breadth=db,
            test_thm_tab=tttab,   bound=b,
            case_breadth=cb,      iterations=i,
            type_range_bound=trb, type_candicates=tc,prelude=p,wrapper=w,toString=s,setupCode=sc,
            cod_term_tab=ctt,     abstest_data_tab=tdt,
		dataconvCode=dc,
            test_data_tab=tjt})
        =
       (Testenv{depth=depth,     breadth=db,
                test_thm_tab=tttab,   bound=b,
                case_breadth=cb,      iterations=i,
                type_range_bound=trb, type_candicates=tc,prelude=p,wrapper=w,toString=s,setupCode=sc,
                cod_term_tab=ctt,     abstest_data_tab=tdt,
		dataconvCode=dc,
                test_data_tab=tjt});

fun breadth_update (breadth)
       (Testenv{depth=dd, breadth=db,
            test_thm_tab=tttab,   bound=b,
            case_breadth=cb,      iterations=i,
            type_range_bound=trb, type_candicates=tc,prelude=p,wrapper=w,toString=s,setupCode=sc,
            cod_term_tab=ctt,     abstest_data_tab=tdt,
		dataconvCode=dc,
            test_data_tab=tjt})
        =
       (Testenv{depth=dd,     breadth=breadth,
                test_thm_tab=tttab,   bound=b,
                case_breadth=cb,      iterations=i,
                type_range_bound=trb, type_candicates=tc,prelude=p,wrapper=w,toString=s,setupCode=sc,
                cod_term_tab=ctt,     abstest_data_tab=tdt,
		dataconvCode=dc,
                test_data_tab=tjt});

fun bound_update (bound)
       (Testenv{depth=dd, breadth=db,
                test_thm_tab=tttab,   bound=b,
                case_breadth=cb,      iterations=i,
                type_range_bound=trb, type_candicates=tc,prelude=p,wrapper=w,toString=s,setupCode=sc,
                cod_term_tab=ctt,     abstest_data_tab=tdt,
		dataconvCode=dc,
                test_data_tab=tjt})
        =
       (Testenv{depth=dd,     breadth=db,
                test_thm_tab=tttab,   bound=bound,
                case_breadth=cb,      iterations=i,
                type_range_bound=trb, type_candicates=tc,prelude=p,wrapper=w,toString=s,setupCode=sc,
                cod_term_tab=ctt,     abstest_data_tab=tdt,
		dataconvCode=dc,
                test_data_tab=tjt});

fun case_breadth_update (case_breadth)
       (Testenv{depth=dd, breadth=db,
                test_thm_tab=tttab,   bound=b,
                case_breadth=cb,      iterations=i,
                type_range_bound=trb, type_candicates=tc,prelude=p,wrapper=w,toString=s,setupCode=sc,
                cod_term_tab=ctt,     abstest_data_tab=tdt,
		dataconvCode=dc,
                test_data_tab=tjt})
        =
       (Testenv{depth=dd,     breadth=db,
                test_thm_tab=tttab,   bound=b,
                case_breadth=case_breadth,      iterations=i,
                type_range_bound=trb, type_candicates=tc,prelude=p,wrapper=w,toString=s,setupCode=sc,
                cod_term_tab=ctt,     abstest_data_tab=tdt,
		dataconvCode=dc,
                test_data_tab=tjt});

fun iterations_update (iterations)
       (Testenv{depth=dd, breadth=db,
                test_thm_tab=tttab,   bound=b,
                case_breadth=cb,      iterations=i,
                type_range_bound=trb, type_candicates=tc,prelude=p,wrapper=w,toString=s,setupCode=sc,
                cod_term_tab=ctt,     abstest_data_tab=tdt,
		dataconvCode=dc,
                test_data_tab=tjt})
        =
       (Testenv{depth=dd,     breadth=db,
                test_thm_tab=tttab,   bound=b,
                case_breadth=cb,      iterations=iterations,
                type_range_bound=trb, type_candicates=tc,prelude=p,wrapper=w,toString=s,setupCode=sc,
                cod_term_tab=ctt,     abstest_data_tab=tdt,
		dataconvCode=dc,
                test_data_tab=tjt});

fun type_range_bound_update (type_range_bound)
       (Testenv{depth=dd, breadth=db,
                test_thm_tab=tttab,   bound=b,
                case_breadth=cb,      iterations=i,
                type_range_bound=trb, type_candicates=tc,prelude=p,wrapper=w,toString=s,setupCode=sc,
                cod_term_tab=ctt,     abstest_data_tab=tdt,
		dataconvCode=dc,
                test_data_tab=tjt})
        =
       (Testenv{depth=dd,     breadth=db,
                test_thm_tab=tttab,   bound=b,
                case_breadth=cb,      iterations=i,
                type_range_bound=type_range_bound,type_candicates=tc,
                prelude=p,            wrapper=w,toString=s,setupCode=sc,
                cod_term_tab=ctt,     abstest_data_tab=tdt,
		dataconvCode=dc,
                test_data_tab=tjt});

fun type_candiates_update (type_canditates)
       (Testenv{depth=dd, breadth=db,
                test_thm_tab=tttab,   bound=b,
                case_breadth=cb,      iterations=i,
                type_range_bound=trb, type_candicates=tc,prelude=p,wrapper=w,toString=s,setupCode=sc,
                cod_term_tab=ctt,     abstest_data_tab=tdt,
		dataconvCode=dc,
                test_data_tab=tjt})
        =
       (Testenv{depth=dd,     breadth=db,
                test_thm_tab=tttab,   bound=b,
                case_breadth=cb,      iterations=i,
                type_range_bound=trb, type_candicates=type_canditates,
                prelude=p,wrapper=w,toString=s,setupCode=sc,
                cod_term_tab=ctt,     abstest_data_tab=tdt,
		dataconvCode=dc,
                test_data_tab=tjt});


fun gen_prelude_update (pre)
       (Testenv{depth=dd, breadth=db,
                test_thm_tab=tttab,   bound=b,
                case_breadth=cb,      iterations=i,
                type_range_bound=trb, type_candicates=tc,prelude=p,wrapper=w,toString=s,setupCode=sc,
                cod_term_tab=ctt,     abstest_data_tab=tdt,
		dataconvCode=dc,
                test_data_tab=tjt})
        =
       (Testenv{depth=dd,     breadth=db,
                test_thm_tab=tttab,   bound=b,
                case_breadth=cb,      iterations=i,
                type_range_bound=trb, type_candicates=tc,prelude=pre,wrapper=w,toString=s,setupCode=sc,
                cod_term_tab=ctt,     abstest_data_tab=tdt,
		dataconvCode=dc,
                test_data_tab=tjt});

fun gen_wrapper_update (wrap)
       (Testenv{depth=dd, breadth=db,
                test_thm_tab=tttab,   bound=b,
                case_breadth=cb,      iterations=i,
                type_range_bound=trb, type_candicates=tc,prelude=p,wrapper=w,
		toString=s,setupCode=sc,
		dataconvCode=dc,
                cod_term_tab=ctt,     abstest_data_tab=tdt,
                test_data_tab=tjt})
        =
       (Testenv{depth=dd,     breadth=db,
                test_thm_tab=tttab,   bound=b,
                case_breadth=cb,      iterations=i,
                type_range_bound=trb, type_candicates=tc,prelude=p,
		wrapper=wrap, toString=s,setupCode=sc,
                cod_term_tab=ctt,     abstest_data_tab=tdt,
		dataconvCode=dc,
                test_data_tab=tjt});


fun gen_toString_update (str)
       (Testenv{depth=dd, breadth=db,
                test_thm_tab=tttab,   bound=b,
                case_breadth=cb,      iterations=i,
                type_range_bound=trb, type_candicates=tc,prelude=p,wrapper=w,
		toString=s,setupCode=sc,
		dataconvCode=dc,
                cod_term_tab=ctt,     abstest_data_tab=tdt,
                test_data_tab=tjt})
        =
       (Testenv{depth=dd,     breadth=db,
                test_thm_tab=tttab,   bound=b,
                case_breadth=cb,      iterations=i,
                type_range_bound=trb, type_candicates=tc,prelude=p,
		wrapper=w, toString=str,setupCode=sc,
                cod_term_tab=ctt,     abstest_data_tab=tdt,
		dataconvCode=dc,
                test_data_tab=tjt});


fun setup_code_update (str)
       (Testenv{depth=dd, breadth=db,
                test_thm_tab=tttab,   bound=b,
                case_breadth=cb,      iterations=i,
                type_range_bound=trb, type_candicates=tc,prelude=p,wrapper=w,
		toString=s,setupCode=sc,
		dataconvCode=dc,
                cod_term_tab=ctt,     abstest_data_tab=tdt,
                test_data_tab=tjt})
        =
       (Testenv{depth=dd,     breadth=db,
                test_thm_tab=tttab,   bound=b,
                case_breadth=cb,      iterations=i,
                type_range_bound=trb, type_candicates=tc,prelude=p,
		wrapper=w, toString=s,setupCode=str,
                cod_term_tab=ctt,     abstest_data_tab=tdt,
		dataconvCode=dc,
                test_data_tab=tjt});

fun dataconv_code_update (str)
       (Testenv{depth=dd, breadth=db,
                test_thm_tab=tttab,   bound=b,
                case_breadth=cb,      iterations=i,
                type_range_bound=trb, type_candicates=tc,prelude=p,wrapper=w,
		toString=s,setupCode=sc,
		dataconvCode=dc,
                cod_term_tab=ctt,     abstest_data_tab=tdt,
                test_data_tab=tjt})
        =
       (Testenv{depth=dd,     breadth=db,
                test_thm_tab=tttab,   bound=b,
                case_breadth=cb,      iterations=i,
                type_range_bound=trb, type_candicates=tc,prelude=p,
		wrapper=w, toString=s,dataconvCode=str,
		setupCode=sc,
                cod_term_tab=ctt,     abstest_data_tab=tdt,
                test_data_tab=tjt});






fun get_depth (Testenv{depth=dd, ...}) = dd;

fun get_breadth (Testenv{breadth=db, ...}) = db;

fun get_gen_prelude (Testenv{prelude=p, ...}) = p;
fun get_gen_wrapper (Testenv{wrapper=w, ...}) = w;
fun get_gen_toString (Testenv{toString=w, ...}) = w;
fun get_setup_code (Testenv{setupCode=w, ...}) = w;
fun get_dataconv_code (Testenv{dataconvCode=w, ...}) = w;





fun gen_test_data thy clasimp name te =
    let val bound = #type_range_bound(rep_testenv te)
        val candTs= #type_candicates(rep_testenv te)
        val tthm  = (the(Symtab.lookup(thm_tab te)name)
	             handle Option => error("No test theorem: "^name^" available!"))
        val atdata= (the(Symtab.lookup(absdata_tab te)name)
                     handle Option => [])
        val cdtab = #cod_term_tab(rep_testenv te)
        val ss    = snd clasimp
        val iters = #iterations(rep_testenv te)
        val jdmts = gen_test_data_core thy name cdtab ss iters 
                                       bound candTs tthm atdata
    in  jdmt_tab_update 
           (te)
           (Symtab.update(name,jdmts)(jdmt_tab te))
    end;


fun get_test_data (te,name) =
    Library.flat(map (convert_goals_to_metahyps (not o is_thyp))
             (the(Symtab.lookup(jdmt_tab te)name)
	      handle Option => error("No data statement: "^name^" available!")));


fun get_test_hyps (te,name) =
    Library.flat(map (convert_goals_to_metahyps (is_thyp))
             (the(Symtab.lookup(jdmt_tab te)name)
	      handle Option => error("No data statement: "^name^" available!")));


(* ************************************************************************* *)
(* micellaneous functions ...                                                *)
(* ************************************************************************* *)


fun animate thy (name, te) =  (* HACK - very simplistic implementation *)
    let val h = !trace_simp;
        fun animate_data_statement thm = 
                (writeln ">>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>";
                 Display.print_thm thm;
                 writeln "============================================";
   (*              asm_full_simplify (simpset_of(theory_of_thm(thm))) thm; *)
                 asm_full_simplify (simpset_of(thy)) thm; 
                 writeln "<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<");
    in trace_simp:=true; 
       map animate_data_statement (get_test_data(te,name));
       trace_simp:=h
    end;



fun export_test_data thy filename dataname te = let
    val thms        = PureThy.get_thms thy (dataname^".test_data")
    val margin = ref 78;
    (*    val thmsstrings = Pretty.string_of (Display.pretty_thms thms) *)
    val today = (Date.toString(Date.fromTimeUniv (Time.now())))^" (UTC)";
    val thyname = Context.theory_name thy
    fun generate (thmlist:(thm list)) = let
	val num = Int.toString(length thms) 
	val thmsstrings = String.concat (map (fn t => "<testcase>\n"
						    ^((Display.string_of_cterm o cprop_of) t)^"\n"
						    ^"</testcase>\n"
						    ) thms)
    in
	"<!---------------------------------------------------------------------\n"
	^" --        Test-Data ("^thyname^"."^dataname^", "^num^" test cases)\n"
	^" --           generated by HOL-TestGen "^testgen_version^"\n"
	^" --           generated on "^today^"\n"
	^" --------------------------------------------------------------------->\n\n"
	^thmsstrings
    end
					
    val test_data_str  = setmp print_mode [] (Pretty.setmp_margin (!margin) (generate)) thms
    val _  = File.write (Path.explode filename) test_data_str;	

in
    thy
end

fun gen_testscript thy put smlput fname dataname te =  let
    val margin = ref 58;
    fun str2term (sign:theory) (s:string) = (Thm.term_of o Thm.read_cterm sign ) 
						 (rpair (TypeInfer.anyT []) s)
    val put_term = str2term ( thy) put
    val put_is_const = is_Const put_term 
    val long_put = if put_is_const 
		   then let 
			   val Const(s,_) = put_term
		       in 
			   s
		       end
		   else put

    fun output_code gr xs = implode (map (snd o Graph.get_node gr)
					 (rev (Graph.all_preds gr xs)));
	

    fun output_code gr  xs =
	let
	    val module = ""
	    val code = List.mapPartial (fn s =>
					   let val c as (_, module', _) = Graph.get_node gr s
					   in if module = "" orelse module = module' then SOME (s, c) else NONE end)
				       (rev (Graph.all_preds gr xs));
	in
	    implode (map (#3 o snd) code)
	end;
	

    fun gen_generate_tc_code prep_term num thy  (speclist: (term list * term) list)  =  let 
	val numstr = Int.toString(num)
	fun image_type typ = (case typ of
				  Type(s,[t]) =>  Type(s,[t])
				| Type("fun",(t::tl)) => image_type (hd (rev tl)) 
				| t => t);    
	val sg =  thy;
	val emptygr : Codegen.codegr = (Graph.empty,
				((Symtab.empty, Symtab.empty), (Symtab.empty, Symtab.empty)));

	val gen_wrapper = get_gen_wrapper te;
	val gen_toString = get_gen_toString te;
	val initGraph = Codegen.new_node ("<Top>", (NONE, "", "")) emptygr
	val gr = initGraph
	val smlput = if (smlput ="") then put else smlput
						   
	fun generate (speclist: (term list * term) list) = let
	    fun arity_of_type typ = (case typ of
					 Type(s,[]) =>  0
				       | Type("fun",(t::tl)) => 1 + arity_of_type (hd tl)
				       | _ => 0
				    )
				    
	    fun arity_of src t = 
		case t of 
		    (Const(s,T)) => if (s=src) then (arity_of_type T) else 0
		  | (Free(s,T)) => if (s=src) then (arity_of_type T) else 0
		  | (Var(i,T)) => 0
		  | (Bound(i)) => 0
		  | (Abs(a,T,t)) => arity_of src t
		  | (f$t) => Int.max ((arity_of src f),(arity_of src t))
			     
	    val put_arity = if (gen_wrapper)
			    then foldl Int.max 0 (map (arity_of long_put) 
			                              (Library.flat (map(fn(xs,y)=>y::xs) 
						                    speclist))) 
			    else 0
	    fun wrap src t = 
		case t of 
		    (Const(s,T)) 
		    => if (s=src) 
		       then if(gen_wrapper)
			    then Free("eval",T)
			    else Free(smlput,T) 
		       else Const(s,T)
		  | (Free(s,T)) 
		    =>  if (s=src)  
			then if(gen_wrapper)
			     then Free("eval",T)
			     else Free(smlput,T)
			else Free(s,T)
		  | (Var(i,T)) => Var(i,T)
		  | (Bound(i)) => Bound(i)
		  | (Abs(a,T,t)) => Abs(a,T,wrap (src) t)
		  | (f$t) => (wrap (src) f )$ (wrap (src) t)
		  | t => t

	    fun gencode (gr, (t:term)) = Codegen.invoke_codegen thy (Codegen.mk_deftab thy) "<Top>" ""
								    false (gr, (wrap long_put t))

          fun to_code gr termlist = foldl_map (gencode)  (gr, map (prep_term sg) termlist);
	  fun codeToString clist = (map (fn p => Pretty.string_of 
						     (Pretty.block [Pretty.brk 1, p])) clist);
	    fun code_spec (gr,(pre,post)) = let
		val (gr_pre, ps_pre) = to_code gr pre;
		val ps_pre = codeToString ps_pre
		val (gr_post,ps_post) = to_code gr_pre [post];
		val ps_post = codeToString ps_post
		    
	    in
		(gr_post, (ps_pre, hd ps_post))
	    end			    

	    val (gr', code) = foldl_map (code_spec) (gr, speclist)
	
	    fun check_code (num,(pre,post)) = let 
		val numstr = Int.toString num;
		val num = num +1;
	    in
		(num, "val _ = print (\"\\nRunning Test Case "^numstr^":\\n\")\n"
                      ^"val pre_"^numstr
		      ^"  = [" ^
		      (space_implode ",\n             " 
				     (map (fn s => "fn () => (" ^ s ^ ")") pre)
		      ) 
		      ^ "];\n" ^
		      "val post_"^numstr^" = " ^
		      "fn () => (" ^ post ^ ")"
		      ^ ";\n"^
		      "val res_"^numstr^"  = TestHarness.check retval pre_"^numstr^
		      " post_"^numstr^";\n"^
		      "val testres = testres@[res_"^numstr^"];\n\n"
		)
	    end

	    val prelude_str = if(get_gen_prelude te) 
			      then   (output_code (fst gr')   ["<Top>"] )
			      else ""
	    val (num,check_str)   = foldl_map (check_code) (num,code);
	    fun toString typ = (case typ of 
				    Type ("boolN", [])         => "Bool.toString"
				  | Type ("Int.int", [])  => "Int.toString"
				  | Type("Real.real", []) =>" Real.toString"
			 	  | Type("List.list",[T]) => 
				    "(fn l => (\"[\"^(foldr (fn (s1,s2) \n"^
				    "                 => (if s2 = \"\" then s1 else s1^\", \"^s2))\n"^
				    "                    \"\" (map (fn x => "^(toString T)^
				    " x) l))^\"]\")) " 
				  | _ => "(fn s => \"\") ")
	    fun smltype_of typ = (case typ of 
				    Type ("boolN", [])       => SOME "bool"
				  | Type ("Int.int", [])  => SOME "int"
				  | Type("RealDef.real", []) => SOME "real"
			 	  | Type("List.list",[T]) => (case smltype_of T of
				  	                           SOME t => SOME ("("^t^" list)")
							         | NONE => NONE)
				  | _ => NONE
			      ) 
	    val wrap_str = (if (gen_wrapper) 
			   then let
				   val cdtab = #cod_term_tab(rep_testenv te)   
				   val putT = if (put_is_const) 
					      then image_type (type_of (str2term ( thy) put))
					      else 
						  let 
						      fun isput (Free(s,t)) = if (s=put) 
									    then true 
									    else false
							| isput _ = false 
						      fun frees (a,b) = term_frees b
						      val postF = filter (isput) 
									 (Library.flat (map (frees) 
										    speclist)); 
						      val Free(_,freetype) = hd postF
						  in 
						      image_type freetype
						  end
						  
				   val retval_term = random_term thy cdtab putT 
				   val (gr,ps) = to_code initGraph (* (Graph.new_node ("<Top>", (NONE, "")) 
									 Graph.empty) *) [retval_term];
				   val retval = hd (codeToString ps);
				   	      
						      
				   fun vars n =  if(n = 0) then "" 
						 else " x"^Int.toString(n)^" "^(vars (n-1))
				   val toStrings = if(gen_toString = "") 
						   then "fun toString  a = " ^(toString putT)^" a;"
						   else "";

				   val retvaltype = case (smltype_of putT) of
				   	              SOME t => ":("^t^")"
						    | NONE => ""
			       in
				   "val return = ref ("^retval^retvaltype^");\n"^
				   "fun eval "^(vars (put_arity))^" =  let val ret = "^smlput^" "
				   ^(vars (put_arity))^" in ((return := ret);ret)end\n"  ^
				   "fun retval () = SOME(!return);\n" ^
				   toStrings^"\n\n" 
			       end 
			   else
			       let
				   val toStrings = if(gen_toString = "") 
						   then  "fun toString a = \"\";"
						   else "";
			       in
				   "fun retval () = NONE;\n"^
				   toStrings^"\n\n"
 			       end)
			       
	 	val toStringName = if (gen_toString = "") then "toString" else gen_toString;
		    
	in  
	    
	    "(*******************************************************************\n"
	    ^" *                         Test-Driver \n"
			^" *           generated by HOL-TestGen "^testgen_version^"\n"
	    ^" *******************************************************************)\n\n"
	    ^"structure TestDriver : sig end = struct\n\n"
	    ^(get_setup_code te)^"\n\n"
	    ^prelude_str
	    ^(get_dataconv_code te)^"\n\n"
	    ^ wrap_str
	    ^ "val testres = [];\n\n"
	    ^ (foldl (fn (s1,s2) => s1^s2) "" check_str)
	    ^ "val _ = TestHarness.printList "^toStringName^" testres;\n\n"
	    ^ "end\n"
	end
    in		  
	setmp print_mode [] (Pretty.setmp_margin (!margin) (generate)) speclist
    end; 
	
	
    val generate_tc_code_i =  gen_generate_tc_code (K I); 
	
	
    val td = get_test_data (te,dataname);
    val prop_terms = map (fn p => let val _ $ ret = concl_of p  in ret end) td;  
    val prems_terms = map (fn data => (
			      map (fn p => let val _ $ ret = p in ret end) (prems_of data))) td;
	
    val spec = map2 (fn x => fn y => (x,y))  prems_terms prop_terms
    val tc   = generate_tc_code_i 0 thy spec;	   
    val _    = File.write (Path.explode fname) tc;
in
    thy
end;

    
end (*struct *);
	  

(* Embedding into Theory Data : *)

structure TestGenData : THEORY_DATA_ARGS =
struct
  val name = "testenv" 
  type T = TestGen.testenv
  val empty = TestGen.mt_testenv
  fun copy T = T
  fun extend T = T
  val merge = fn _ => TestGen.merge_testenv
  fun print sg te =
      let val {depth,breadth,test_thm_tab,
              bound,case_breadth,iterations,type_range_bound,
              type_candicates,prelude,wrapper,toString,setupCode,dataconvCode,cod_term_tab, abstest_data_tab,test_data_tab} =
              TestGen.rep_testenv te
          fun H (n,thm) = (writeln (n^":");Display.print_thm thm)
          fun H2(n,thms)= (writeln (n^":");
                       writeln "======";
                       map (fn x => Display.print_thm x) thms;
                       writeln "======") 
  in  (writeln ">>> Testenvironment >>>>>>>>>>>>>>>>>";
       writeln "+++ Control Data: +++++++++++++++++++";
       writeln("*** default depth:   "^Int.toString(depth));
       writeln("*** default breadth: "^Int.toString(depth));
       writeln("*** bound:           "^Int.toString(bound));
       writeln("*** iterations:      "^Int.toString(iterations));
       writeln "+++ Testtheoremtable: +++++++++++++++";
       map H (Symtab.dest test_thm_tab);
       writeln "+++ Testjudgements: +++++++++++++++++";
       map H2 (Symtab.dest test_data_tab);
       writeln "+++ Testdatatable: ++++++++++++++++++";
       map H2 (Symtab.dest abstest_data_tab);
       writeln "<<< Testenvironment <<<<<<<<<<<<<<<<<")
  end
end;

structure TestGen_DataManagement = GenericDataFun(TestGenData);
(* val TestGen_setup = TestGen_DataManagement.init *) 
