(*****************************************************************************
 * HOL-TestGen --- theorem-prover based test case generation
 *                 http://www.brucker.ch/projects/hol-testgen/
 *                                                                            
 * Testing.thy --- the base testing theory
 * This file is part of HOL-TestGen.
 *
 * Copyright (c) 2005-2007, ETH Zurich, Switzerland
 *               2008 Achim D. Brucker, Germany 
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *
 *     * Neither the name of the copyright holders nor the names of its
 *       contributors may be used to endorse or promote products derived
 *       from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 ******************************************************************************)
(* $Id:$ *)

header {* Basic Testing Setup *}

theory Testing imports Main
  uses  "kernel_ext/isabelle2008_kernel_patch" "TestGen.ML" "isar_setup.ML" begin 

(* setup "TestGen_setup" *)

setup "TestingIsar_setup"

section{*General Framework for Unit-Test*}

constdefs         THYP   :: "bool \<Rightarrow> bool"
                 "THYP x \<equiv> x"
		  RSF    :: "bool"
		 "RSF    \<equiv> True" (* just to denote random solver failures *)

lemma  THYP_triv [simp]: "THYP True = True" by(simp add: THYP_def)

lemma  THYP_I    : "A \<Longrightarrow> THYP A" by(simp add: THYP_def)

lemma  THYP_D    : "THYP A \<Longrightarrow> A" by(simp add: THYP_def) 

lemma  THYP_E    : "\<lbrakk> THYP A; (A \<Longrightarrow> R) \<rbrakk> \<Longrightarrow> R" by(simp add: THYP_def)

lemma  THYP_spec : "THYP(\<forall>x. P x) \<Longrightarrow> THYP(P x)" by(simp add: THYP_def)

lemma  THYP_app1 : "\<lbrakk> THYP(A \<longrightarrow> B); A \<rbrakk> \<Longrightarrow> B" by(simp add: THYP_def) 
       (* to be used with etac in order to apply uniformity hyp *)

lemma  THYP_app1_rev : "\<lbrakk> A; THYP(A \<longrightarrow> B) \<rbrakk> \<Longrightarrow> B" by(simp add: THYP_def) 
       (* to be used with etac in order to apply uniformity hyp *)

lemma  THYP_app2 : "\<lbrakk> THYP(A); A \<Longrightarrow> B \<rbrakk> \<Longrightarrow> B" by(simp add: THYP_def) 
       (* to be used with RS in order to apply regularity hyp *)


lemma  THYP_E_reg :  "\<lbrakk> size x \<le> k \<Longrightarrow> B; THYP((k < size x) \<longrightarrow> B)\<rbrakk> \<Longrightarrow> B" 
       by(simp add: THYP_def, auto)
       (* to be used with rtac  in order to introduce regularity hyp *)


lemma  RSF_E : "\<lbrakk> RSF \<Longrightarrow> P \<rbrakk> \<Longrightarrow> P"
       by(auto simp: RSF_def)
consts_code  "RSF" ("TestHarness.rsf")


section {* ISAR Interface Setup *}

text {* Method for generating test cases *}
method_setup gen_test_cases = {*
let 
  fun gtc args ctx =  
  let 
     val _               = (spec_time := Timer.checkRealTimer (!tc_timer))
     val thy             = ProofContext.theory_of ctx;
     val dm              = TestGen_DataManagement.get(Context.Theory thy);
     val default_depth   = TestGen.get_depth dm;
     val default_breadth = TestGen.get_breadth dm;
     fun clasimp_meth tac prems ctxt = Method.METHOD (fn facts =>
                                         ALLGOALS (Method.insert_tac (prems @ facts)) 
                                                THEN tac (local_clasimpset_of ctxt));
     fun clasimp_meth' tac prems ctxt = Method.METHOD (fn facts =>
                                         HEADGOAL (Method.insert_tac (prems @ facts) 
                                                THEN' tac (local_clasimpset_of ctxt)));
     val clasimp_method' = Method.bang_sectioned_args clasimp_modifiers o clasimp_meth';
     fun gen_test_case_args m = Method.bang_sectioned_args' clasimp_modifiers 
                                  (Scan.lift (Scan.option (Args.nat -- Args.nat) -- Args.name)) m;
     fun gen_test_case_meth (NONE,prog) = 
                         gen_test_case_meth(SOME(default_depth,default_breadth),prog)
       | gen_test_case_meth ((SOME(m, n),prog)) = clasimp_meth (CHANGED_PROP o 
                         (fn css => TestGen.gen_test_case_tac (thy) css m n prog));
  in 
      gen_test_case_args gen_test_case_meth args ctx
  end

in 
   gtc
end

*} "generate symbolic test cases"


(* setup setup_TestGen *)

section{*General Framework for Sequence-Test*}

text{* 
  It is necessary to model the dynamic behaviour of our
  specification. This can be done using a fold operator which takes a
  history, a state, and a state-transition and returns a new
  state. With this operator we get the resulting state after applying
  a sequence of state transitions induced by a list of incoming
  packets to the initial state.  
*}
consts Mfold :: "'\<iota> list  \<Rightarrow>  '\<sigma> \<Rightarrow>  ('\<iota> \<Rightarrow> '\<sigma> \<rightharpoonup> '\<sigma>) \<rightharpoonup> '\<sigma>"  
primrec
  "Mfold [] S ST = Some S"
  "Mfold (a#H) S ST = (case ST a S of  None    \<Rightarrow>   None
                                     | Some Sn \<Rightarrow>   Mfold H Sn ST)"

text{* The following adaption combinator converts an input-output program
   under test of type: $'\<iota> \<Rightarrow> '\<sigma> \<rightharpoonup> 'o \<times>'\<sigma>$ with program state $\sigma$ 
   into a state transition program that can be processed by 
   Mfold. The key idea to turn Mfold into a test-driver for
   a reactive system is by providing an internal state $\sigma'$,
   managed by the test driver, and external, problem-specific
   functions "rebind" and "substitute" that operate on this
   internal state. For example, this internal state can be
   instantiated with an environment $var \<rightharpoonup> value$. The output
   (or parts of it) can then be bound to vars in the environment.
   In contrast, substitute can then explicit variables occuring
   in value representations into pure values, e.g. is can substitue
   $c~("X")$ into $c~3$ provided the environment contained the map
   with $X \leadsto 3$. *}  

text{* The state of the test-driver consists of two parts: the state
   of the observer (or: adaptor) $\sigma$ and the internal state $\sigma'$ of the
   the step-function of the system under test $ioprog$ is allowed to use. *} 

constdefs observer :: "['\<sigma> \<Rightarrow> 'o \<Rightarrow> '\<sigma>, '\<sigma> \<Rightarrow> '\<iota> \<Rightarrow> '\<iota>, '\<sigma>\<times>'\<sigma>' \<Rightarrow> '\<iota> \<Rightarrow> 'o \<Rightarrow> bool] 
                       \<Rightarrow>  ('\<iota> \<Rightarrow> '\<sigma>' \<rightharpoonup> 'o \<times>'\<sigma>') 
                       \<Rightarrow>  ('\<iota> \<Rightarrow> ('\<sigma>\<times>'\<sigma>' \<rightharpoonup> '\<sigma>\<times>'\<sigma>'))"

         "observer rebind substitute postcond ioprog  \<equiv>  
          (\<lambda> input. (\<lambda> (\<sigma>, \<sigma>'). let input'= substitute \<sigma> input in 
                                 case ioprog input' \<sigma>' of
                                    None \<Rightarrow> None  (* ioprog failure - eg. timeout ... *)  
                                  | Some (output, \<sigma>''') \<Rightarrow> let \<sigma>'' = rebind \<sigma> output in
                                                          (if postcond (\<sigma>'',\<sigma>''') input' output
                                                          then Some(\<sigma>'', \<sigma>''')
                                                          else None (* postcond failure *) )))"  

end
